import os
import shlex
import json
import tests
from io import StringIO
from pathlib import Path
from unittest.mock import patch
from e4s_cl import config
from e4s_cl.variables import set_dry_run
from e4s_cl.cf.launchers import LAUNCHERS
from e4s_cl.cli.commands.launch import COMMAND
from e4s_cl.cli.cli_view import CreateCommand
from e4s_cl.model.profile import Profile


class LaunchTest(tests.TestCase):

    @classmethod
    def setUpClass(cls):
        tests.TestCase.setUpClass()
        CONFIGURATION_FILE = tests.ASSETS / "e4s-cl.yaml"
        TEST_CONFIGURATION = config.Configuration.create_from_file(
            CONFIGURATION_FILE)
        config.update_configuration(TEST_CONFIGURATION)

    @classmethod
    def tearDownClass(cls):
        config.update_configuration(config.Configuration.default())

    def setUp(self):
        self.resetStorage()

        data = os.getenv('__E4S_CL_TEST_PROFILE', {})
        if not data:
            return

        controller = Profile.controller()
        profile = controller.create(json.loads(data))
        controller.select(profile)

    def tearDown(self):
        self.resetStorage()


def wrapper(launcher):

    def generated(self):
        set_dry_run(True)
        self.assertCommandReturnValue(
            0, COMMAND,
            shlex.split(
                f"--backend containerless --image None {launcher} hostname"))

    generated.__name__ = f"test_launch_{launcher}"

    return generated


for launcher in LAUNCHERS:
    test = wrapper(launcher)
    setattr(LaunchTest, test.__name__, test)



"""
Ensure the profile unselect command behaves properly
"""
import tests
from e4s_cl.model.profile import Profile
from e4s_cl.error import ProfileSelectionError
from e4s_cl.cli.commands.profile.unselect import COMMAND


class ProfileUnselectTest(tests.TestCase):
    """
    Profiles are erased after each tests
    """
    def setUp(self):
        self.profile_name = 'test_unselect_profile'
        profile = Profile.controller().create({'name': self.profile_name})
        Profile.controller().select(profile)

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_unselect(self):
        self.assertCommandReturnValue(0, COMMAND, [self.profile_name])
        self.assertRaises(ProfileSelectionError, Profile.controller().selected)

    def test_unselect_nonexistent(self):
        self.assertCommandReturnValue(2, COMMAND, ["non-existent"])



from pathlib import Path
import sys
import shlex
import tests
from tempfile import TemporaryFile
from e4s_cl.cf.containers import Container, BackendError

__IMAGE__ = "centos:7"

try:
    import docker
    from e4s_cl.cf.containers.docker import DockerContainer
    try:
        client = docker.from_env()
        try:
            client.images.get(__IMAGE__)
        except BackendError as err:
            IMAGE_AVAIL = False
        else:
            IMAGE_AVAIL = True
        DOCKER_AVAIL = True
    except docker.errors.DockerException:
        raise ModuleNotFoundError
except ModuleNotFoundError:
    DOCKER_AVAIL = False
    IMAGE_AVAIL = False


@tests.skipUnless(DOCKER_AVAIL, "Failed to connect to the docker daemon")
class DockerContainerTest(tests.TestCase):

    def test_create(self):
        container = Container(name='docker', image=__IMAGE__)
        self.assertFalse(type(container) == Container)
        self.assertTrue(type(container) == DockerContainer)
        self.assertTrue(isinstance(container, Container))

    @tests.skipUnless(IMAGE_AVAIL, f"Image {__IMAGE__} not available")
    def test_run(self):
        container = Container(name='docker', image=__IMAGE__)
        container.run(shlex.split("ls"))

    @tests.skipUnless(IMAGE_AVAIL, f"Image {__IMAGE__} not available")
    def test_run_bound_files(self):
        container = Container(name='docker', image=__IMAGE__)

        returncode = container.run(shlex.split("ls /e4s-cl/tests"))
        self.assertNotEqual(returncode, 0)

        container.bind_file(Path(__file__).parent, dest='/e4s-cl/tests')
        returncode = container.run(shlex.split("ls /e4s-cl/tests"))

        self.assertEqual(returncode, 0)

    @tests.skipUnless(IMAGE_AVAIL, f"Image {__IMAGE__} not available")
    def test_run_bound_env(self):
        container = Container(name='docker', image=__IMAGE__)
        bkpstream = sys.stdout
        sys.stdout = TemporaryFile(mode="w+")

        container.bind_env_var("MYVAR", "MYVALUE")
        container.run(shlex.split("env"))
        sys.stdout.seek(0, 0)
        self.assertIn("MYVALUE", sys.stdout.read())

        sys.stdout = bkpstream

    def test_run_bad_image(self):
        container = Container(name='docker', image='sike_not_an_image')
        with self.assertRaises(BackendError):
            container.run(shlex.split("ls"))



import tests
import shlex
from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
from e4s_cl.cf.detect_mpi import MPIIdentifier
from e4s_cl.cf.wi4mpi import (wi4mpi_adapt_arguments)
from e4s_cl.cf.wi4mpi.install import (WI4MPI_RELEASE_URL, _download_wi4mpi,
                                      _update_config)


class Wi4MPITest(tests.TestCase):

    def test_update_config(self):
        with NamedTemporaryFile(mode='w', delete=False) as config:
            config_file = config.name
            config.write("MPICH_DEFAULT_ROOT=\"path/to/installation\"")

        _update_config(config_file, 'MPICH_DEFAULT_ROOT', 'new/value')
        _update_config(config_file, 'OPENMPI_DEFAULT_ROOT', 'other/value')

        with open(config_file, 'r') as config:
            data = config.read()

            self.assertIn('MPICH_DEFAULT_ROOT="new/value"', data)
            self.assertIn('OPENMPI_DEFAULT_ROOT="other/value"', data)

        Path(config_file).unlink()

    def test_adapt_arguments(self):
        command = shlex.split(
            "mpirun -np 4 -T mpich -F openmpi --mca btl_tcp_if_include ib0 -x PATH"
        )
        converted = wi4mpi_adapt_arguments(command)
        self.assertEqual(
            converted,
            shlex.split(
                "mpirun -T mpich -F openmpi -E \"-np 4 --mca btl_tcp_if_include ib0 -x PATH\""
            ))

        command = shlex.split("mpirun -T mpich -F openmpi")
        converted = wi4mpi_adapt_arguments(command)
        self.assertEqual(converted, shlex.split("mpirun -T mpich -F openmpi"))

        command = shlex.split("mpirun -np 4 --mca param value")
        converted = wi4mpi_adapt_arguments(command)
        self.assertEqual(converted,
                         shlex.split("mpirun -E '-np 4 --mca param value'"))



"""
Remove the selected status from the selected profile.
"""

from e4s_cl import EXIT_SUCCESS
from e4s_cl.cli import arguments
from e4s_cl.model.profile import Profile
from e4s_cl.cli.command import AbstractCommand


class ProfileUnselectCommand(AbstractCommand):
    """``profile unselect`` subcommand."""
    def _construct_parser(self):
        usage = f"{self.command}"
        parser = arguments.get_model_identifier(Profile,
                                                prog=self.command,
                                                usage=usage,
                                                description=self.summary)
        return parser

    def main(self, argv):
        profile_name = self._parse_args(argv).profile.get('name')

        if profile_name != Profile.selected().get('name'):
            self.parser.error(f"Profile {profile_name} is not selected.")

        Profile.controller().unselect()
        return EXIT_SUCCESS


COMMAND = ProfileUnselectCommand(
    __name__,
    summary_fmt=("Unselect the selected profile.\n"
                 "Use `profile list` to see all profiles."))



"""
Module home of the argument parser methods and helpers
"""

import sys
import re
import copy
import argparse
import textwrap
from pathlib import Path
from gettext import gettext as _
from operator import attrgetter
from e4s_cl import logger, util
from e4s_cl.cli import USAGE_FORMAT
from e4s_cl.util import add_dot
from e4s_cl.error import InternalError
from e4s_cl.cf.storage import StorageError
from e4s_cl.cf.storage.levels import ORDERED_LEVELS, STORAGE_LEVELS

Action = argparse.Action
"""Action base class."""

ArgumentError = argparse.ArgumentError
"""Argument error exception base class."""

ArgumentTypeError = argparse.ArgumentTypeError
"""Argument type error exception."""

ArgumentsNamespace = argparse.Namespace
"""Generic container for parsed arguments."""

SUPPRESS = argparse.SUPPRESS
"""Suppress attribute creation in parsed argument namespace."""

REMAINDER = argparse.REMAINDER
"""All the remaining command-line arguments are gathered into a list."""

UNSELECTED = "==NOT_SELECTED=="

STORAGE_LEVEL_FLAG = "@"
"""Command line flag that indicates storage level."""

_DEFAULT_STORAGE_LEVEL = ORDERED_LEVELS[0].name

LOGGER = logger.get_logger(__name__)


class MutableArgumentGroup(argparse._ArgumentGroup):
    """Argument group that allows its actions to be modified after creation."""

    # We're changing the behavior of the superclass so we need to access protected members
    # pylint: disable=protected-access

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, option_string):
        return self._option_string_actions[option_string]


class MutableArgumentGroupParser(argparse.ArgumentParser):
    """Argument parser with mutable groups and better help formatting.

    :py:class:`argparse.ArgumentParser` doesn't allow groups to change once set 
    and generates "scruffy" looking help, so we fix this problems in this subclass.
    """

    # We're changing the behavior of the superclass so we need to access protected members
    # pylint: disable=protected-access

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.actions = self._actions

    def __getitem__(self, option_string):
        return self._option_string_actions[option_string]

    def exit(self, status=0, message=None):
        if status and message:
            LOGGER.error(message)
        sys.exit(status)

    def error(self, message):
        """From the sources of python 3.8:
        error(message: string)

        Prints a usage message incorporating the message to stderr and
        exits.

        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        LOGGER.error(self.format_usage())
        args = {'prog': self.prog, 'message': message}
        self.exit(2, _('%(prog)s: error: %(message)s\n') % args)

    def add_argument_group(self, *args, **kwargs):
        """Returns an argument group.
        
        If the group doesn't exist it will be created.
        
        Args:
            *args: Positional arguments to pass to :any:`ArgumentParser.add_argument_group`
            **kwargs: Keyword arguments to pass to :any:`ArgumentParser.add_argument_group`

        Returns:
            An argument group object.
        """
        title = kwargs.get('title', args[0])
        for group in self._action_groups:
            if group.title == title:
                return group
        group = MutableArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group

    def _format_help_console(self):
        """Format command line help string."""
        formatter = self._get_formatter()
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)
        formatter.add_text(self.description)
        for action_group in self._sorted_groups():
            title = ' '.join(x[0].upper() + x[1:]
                             for x in action_group.title.split())
            formatter.start_section(title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(
                sorted(action_group._group_actions,
                       key=attrgetter('option_strings')))
            formatter.end_section()
        formatter.add_text(self.epilog)
        return formatter.format_help()

    def format_help(self):
        try:
            func = getattr(self, '_format_help_' + USAGE_FORMAT.lower())
        except AttributeError as attr_err:
            raise InternalError(
                f"Invalid USAGE_FORMAT: {USAGE_FORMAT}") from attr_err
        return func()

    def _sorted_groups(self):
        """Iterate over action groups."""
        positional_title = 'positional arguments'
        optional_title = 'optional arguments'
        groups = sorted(self._action_groups, key=lambda x: x.title.lower())
        for group in groups:
            if group.title == positional_title:
                yield group
                break
        for group in groups:
            if group.title == optional_title:
                yield group
                break
        for group in groups:
            if group.title not in [positional_title, optional_title]:
                yield group

    def merge(self,
              parser,
              group_title=None,
              include_positional=False,
              include_optional=True,
              include_storage=False,
              exclude_groups=None,
              exclude_arguments=None):
        """Merge arguments from a parser into this parser.
        
        Modify this parser by adding additional arguments copied from the supplied parser.
        
        Args:
            parser (MutableArgumentGroupParser): Parser to pull arguments from.
            group_title (str): Optional group title for merged arguments.
            include_positional (bool): If True, include positional arguments.
            include_optional (bool): If True, include optional arguments.
            include_storage (bool): If True, include the storage level argument, see :any:`STORAGE_LEVEL_FLAG`.
            exclude_groups (list): Strings identifying argument groups that should be excluded.
            exclude_arguments (list): Strings identifying arguments that should be excluded.
        """
        for action in parser.actions:
            if exclude_groups and action.container.title in exclude_groups:
                continue
            dst_group = self.add_argument_group(
                group_title if group_title else action.container.title)
            optional = bool(action.option_strings)
            storage = '-' + STORAGE_LEVEL_FLAG in action.option_strings
            excluded = exclude_arguments and bool([
                optstr for optstr in action.option_strings
                for substr in exclude_arguments if substr in optstr
            ])
            # pylint: disable=too-many-boolean-expressions
            if (excluded or (not include_storage and storage)
                    or (not include_optional and optional)
                    or (not include_positional and not optional)):
                continue
            try:
                dst_group._add_action(copy.copy(action))
            except argparse.ArgumentError:
                # Argument is already in this parser.
                pass


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Custom help string formatter for argument parser.
    
    Provide proper help message alignment, line width, and formatting.
    Uses console line width (:any:`logger.LINE_WIDTH`) to format help 
    messages appropriately so they don't wrap in strange ways.
    
    Args:
        prog (str): Name of the program.
        indent_increment (int): Number of spaces to indent wrapped lines.
        max_help_position (int): Column on which to begin subsequent lines of wrapped help strings.
        width (int): Maximum help message length before wrapping.
    """

    def __init__(self,
                 prog,
                 indent_increment=2,
                 max_help_position=30,
                 width=None):
        if width is None:
            width = logger.LINE_WIDTH
        super().__init__(prog, indent_increment, max_help_position, width)

    def _split_lines(self, text, width):
        parts = []
        for line in text.splitlines():
            parts.extend(textwrap.wrap(line, width))
        return parts

    def _get_help_string(self, action):
        helpstr = add_dot(action.help)
        helpstr = helpstr[0].upper() + helpstr[1:]
        # Disabled in favour of per-help customization
        # indent = ' ' * self._indent_increment
        # if choices := getattr(action, 'choices', None)
        # helpstr += '\n%s- %s: %s' % (indent, action.metavar,
        # ', '.join(choices))
        # if '%(default)' not in action.help:
        # if action.default is not argparse.SUPPRESS:
        # defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
        # if action.option_strings or action.nargs in defaulting_nargs:
        # if isinstance(action.default, list):
        # default_str = ', '.join(action.default)
        # else:
        # default_str = str(action.default)
        #helpstr += '\n%s' % indent + '- default: %s' % default_str
        return helpstr

    def _format_positional(self, argstr):
        return argstr

    def _format_optional(self, argstr):
        return argstr

    def _format_requred_arg(self, argstr):
        return argstr

    def _format_optional_arg(self, argstr):
        return argstr

    def _format_meta_arg(self, argstr):
        return argstr

    def _format_args(self, action, default_metavar):
        # pylint: disable=consider-using-f-string
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs is None:
            result = self._format_requred_arg('%s' % get_metavar(1))
        elif action.nargs == argparse.OPTIONAL:
            result = self._format_optional_arg('[%s]' % get_metavar(1))
        elif action.nargs == argparse.ZERO_OR_MORE:
            result = self._format_optional_arg('[%s [%s ...]]' %
                                               get_metavar(2))
        elif action.nargs == argparse.ONE_OR_MORE:
            tpl = get_metavar(2)
            result = self._format_requred_arg(
                '%s' % tpl[0]) + self._format_optional_arg(
                    ' [%s ...]' % tpl[1])
        elif action.nargs == argparse.REMAINDER:
            result = self._format_requred_arg('...')
        elif action.nargs == argparse.PARSER:
            result = self._format_requred_arg('%s ...' % get_metavar(1))
        else:
            formats = ['%s' for _ in range(action.nargs)]
            result = ' '.join(formats) % get_metavar(action.nargs)
        return result

    def _format_action_invocation(self, action):
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return self._format_positional(metavar)

        parts = []
        if action.nargs == 0:
            parts.extend(
                self._format_optional(x) for x in action.option_strings)
        else:
            default = action.dest.upper()
            args_string = self._format_args(action, default)
            for option_string in action.option_strings:
                parts.append(
                    f'{self._format_optional(option_string)} {args_string}')
        return ', '.join(parts)


class ConsoleHelpFormatter(HelpFormatter):
    """Custom help string formatter for console output."""

    def start_section(self, heading):
        return super().start_section(util.color_text(heading, attrs=['bold']))

    def add_argument(self, action):
        if action.help is not SUPPRESS:
            get_invocation = self._format_action_invocation
            invocations = [get_invocation(action)]
            for subaction in self._iter_indented_subactions(action):
                invocations.append(get_invocation(subaction))
            invocation_length = max(
                len(util.uncolor_text(s)) for s in invocations)
            action_length = invocation_length + self._current_indent
            self._action_max_length = max(self._action_max_length,
                                          action_length)
            self._add_item(self._format_action, [action])

    def _format_positional(self, argstr):
        return util.color_text(argstr, 'red')

    def _format_optional(self, argstr):
        return util.color_text(argstr, 'red')

    def _format_requred_arg(self, argstr):
        return util.color_text(argstr, 'blue')

    def _format_optional_arg(self, argstr):
        return util.color_text(argstr, 'cyan')

    def _format_action(self, action):
        # pylint: disable=consider-using-f-string
        help_position = min(self._action_max_length + 2,
                            self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)
        action_header_nocolor = util.uncolor_text(action_header)
        if not action.help:
            # No help; start on same line and add a final newline
            tup = self._current_indent, '', action_header
            action_header = '%*s%s\n' % tup
        elif len(action_header_nocolor) <= action_width:
            # Short action name; start on the same line and pad two spaces
            # Adjust length to account for color control chars
            length = action_width + len(action_header) - len(
                action_header_nocolor)
            tup = self._current_indent, '', length, action_header
            action_header = '%*s%-*s  ' % tup
            indent_first = 0
        else:
            # Long action name; start on the next line
            tup = self._current_indent, '', action_header
            action_header = '%*s%s\n' % tup
            indent_first = help_position
        parts = [action_header]
        if action.help:
            help_text = self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            parts.append('%*s%s\n' % (indent_first, '', help_lines[0])) # pylint: disable=E0606
            for line in help_lines[1:]:
                parts.append('%*s%s\n' % (help_position, '', line))
        elif not action_header.endswith('\n'):
            parts.append('\n')
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))
        return self._join_parts(parts)


class ParseBooleanAction(argparse.Action):
    """Argument parser action for boolean values.
    
    Essentially a wrapper around :any:`e4s_cl.util.parse_bool`.
    """

    # pylint: disable=too-few-public-methods

    def __call__(self, parser, namespace, value, unused_option_string=None):
        """Sets the `self.dest` attribute in `namespace` to the parsed value of `value`.
        
        If `value` parses to a boolean via :any:`e4s_cl.util.parse_bool` then the 
        attribute value is that boolean value.
            
        Args:
            parser (str): Argument parser object this group belongs to.
            namespace (object): Namespace to receive parsed value via setattr.
            value (str): Value parsed from the command line/
        """
        try:
            setattr(namespace, self.dest, util.parse_bool(value))
        except TypeError as type_err:
            raise argparse.ArgumentError(
                self, 'Boolean value required') from type_err


def get_parser(prog=None, usage=None, description=None, epilog=None):
    """Builds an argument parser.
    
    The returned argument parser accepts no arguments.
    Use :any:`argparse.ArgumentParser.add_argument` to add arguments.
    
    Args:
        prog (str): Name of the program.
        usage (str): Description of the program's usage.
        description (str): Text to display before the argument help.
        epilog (str): Text to display after the argument help.

    Returns:
        MutableArgumentGroupParser: The customized argument parser object.
    """
    try:
        formatter = getattr(sys.modules[__name__],
                            USAGE_FORMAT.capitalize() + 'HelpFormatter')
    except AttributeError as attr_err:
        raise InternalError(
            f"Invalid USAGE_FORMAT: {USAGE_FORMAT}") from attr_err
    return MutableArgumentGroupParser(prog=prog,
                                      usage=usage,
                                      description=description,
                                      epilog=epilog,
                                      formatter_class=formatter)


def get_parser_from_model(model,
                          use_defaults=True,
                          prog=None,
                          usage=None,
                          description=None,
                          epilog=None):
    """Builds an argument parser from a model's attributes.
    
    The returned argument parser will accept arguments as defined by the model's `argparse` 
    attribute properties, where the arguments to :any:`argparse.ArgumentParser.add_argument` 
    are specified as keyword arguments.
    
    Examples:
        Given this model attribute:
        ::
        
            'openmp': {
                'type': 'boolean', 
                'description': 'application uses OpenMP',
                'default': False, 
                'argparse': {'flags': ('--openmp',),
                             'metavar': 'T/F',
                             'nargs': '?',
                             'const': True,
                             'action': ParseBooleanAction},
            }

        The returned parser will accept the ``--openmp`` flag accepting zero or one arguments 
        with 'T/F' as the metavar.  If ``--openmp`` is omitted the default value of False will
        be used.  If ``--openmp`` is provided with zero arguments, the const value of True will
        be used.  If ``--openmp`` is provided with one argument then the provided argument will
        be passed to a ParseBooleanAction instance to generate a boolean value.  The argument's
        help description will appear as "application uses OpenMP" if the ``--help`` argument is given.
    
    Args:
        model (Model): Model to construct arguments from.
        use_defaults (bool): If True, use the model attribute's default value 
                             as the argument's value if argument is not specified. 
        prog (str): Name of the program.
        usage (str): Description of the program's usage.
        description (str): Text to display before the argument help.
        epilog (str): Text to display after the argument help.

    Returns:
        MutableArgumentGroupParser: The customized argument parser object.        
    """
    parser = get_parser(prog, usage, description, epilog)
    groups = {}
    for attr, props in model.attributes.items():
        try:
            options = dict(props['argparse'])
        except KeyError:
            if 'primary_key' in props:
                options = {'metavar': f'<{model.name.lower()}_{attr}>'}
            else:
                continue
        if use_defaults:
            options['default'] = props.get('default', argparse.SUPPRESS)
        else:
            options['default'] = argparse.SUPPRESS
        try:
            options['help'] = props['description']
        except KeyError:
            pass
        try:
            group_name = options['group'] + ' arguments'
        except KeyError:
            group_name = model.name.lower() + ' arguments'
        else:
            del options['group']
        group = groups.setdefault(group_name,
                                  parser.add_argument_group(group_name))
        try:
            flags = options['flags']
        except KeyError:
            flags = (attr, )
        else:
            del options['flags']
            options['dest'] = attr
        prop_type = props.get('type', 'string')
        if prop_type == 'array':
            if 'nargs' not in options:
                options['nargs'] = '+'
        elif prop_type == 'boolean':
            if 'action' not in options:
                options['action'] = ParseBooleanAction
            if 'nargs' not in options:
                options['nargs'] = '?'
            if 'const' not in options:
                options['const'] = True
            if 'metavar' not in options:
                options['metavar'] = 'T/F'
        else:
            if 'type' not in options:
                options['type'] = {
                    'integer': int,
                    'float': float,
                    'boolean': str,
                    'string': str
                }[prop_type]
        group.add_argument(*flags, **options)
    return parser


def get_model_identifier(model,
                         prog=None,
                         usage=None,
                         description=None,
                         epilog=None):
    """
    Get a parser identifying a model from its key_attribute. If nothing is 
    specified, the model's selected() method is used as default.
    """
    parser = get_parser(prog, usage, description, epilog)

    model_name = model.name.lower()
    key_attr = model.key_attribute

    _default = SUPPRESS
    if 'selected' in dir(model):
        _default = model.selected().get(key_attr, UNSELECTED)

    parser.add_argument(
        model_name,
        nargs='?',
        type=single_defined_object(model, key_attr),
        help=
        f"The target {model_name}. If omitted, defaults to the selected {model_name}",
        default=_default,
        metavar=f"{model_name}_{key_attr}")

    return parser


def add_storage_flag(parser,
                     action,
                     object_name,
                     plural=False,
                     exclusive=True):
    """Add flag to indicate target storage container.
    
    Args:
        parser (MutableArgumentGroupParser): The parser to modify.
        action (str): The action that will be taken by the command, e.g. "delete" or "list"
        object_name (str): The type of object that will be manipulated, e.g. "application" or "measurement"
        plural (bool): Pluralize help message if True.
        exclusive (bool): Only one storage level may be specified if True.
    """
    help_parts = [
        "%s %ss" if plural else "%s the %s", " at the specified storage ",
        "level" if exclusive else "levels"
    ]
    help_str = "".join(help_parts) % (action, object_name)
    nargs = 1 if exclusive else '+'
    choices = [container.name for container in ORDERED_LEVELS]
    parser.add_argument('-' + STORAGE_LEVEL_FLAG,
                        help=help_str,
                        metavar="<level>",
                        nargs=nargs,
                        choices=choices,
                        default=[_DEFAULT_STORAGE_LEVEL])


def parse_storage_flag(args):
    try:
        names = getattr(args, STORAGE_LEVEL_FLAG)
    except AttributeError:
        names = [_DEFAULT_STORAGE_LEVEL]
    return [STORAGE_LEVELS[name] for name in names]


def posix_path(string):
    """Argument type callback.
    Returns a posix-compliant path."""
    if string:
        return Path(string.strip()).as_posix()
    return ''


def posix_path_list(string):
    """Argument type callback.
    Asserts that the string corresponds to a list of paths."""
    return list(map(posix_path, string.split(',')))


def existing_posix_path(string):
    """
    Argument type callback.
    Returns a posix-compliant path if it exists."""
    path = Path(string.strip())

    if not path.exists():
        raise argparse.ArgumentTypeError(
            f"File {path.as_posix()} does not exist")

    return path


def existing_posix_path_list(string):
    """Argument type callback.
    Asserts that the string corresponds to a list of existing paths."""
    return list(map(existing_posix_path, string.split(',')))


def binary_in_path(string):
    """ Argument type callback. Asserts the given string identifies a launcher binary
    on the system. """

    path = util.which(string)
    if not path:
        raise ArgumentTypeError(
            f"Launcher argument '{string}' could not be resolved to a binary")
    return path


def _search_available_databases(model, field, regex):
    matches = []

    for level in ORDERED_LEVELS:
        try:
            matches.extend(model.controller(storage=level).match(field, regex))
        except StorageError:
            LOGGER.debug("Failed to access records from level %s", level.name)

    return matches


def single_defined_object(model, field):
    """Argument type callback.
    Asserts that the string corresponds to an existing object."""

    def wrapper(string):
        if string == UNSELECTED:
            raise argparse.ArgumentTypeError(
                f"no {model.name} selected nor specified")

        matches = _search_available_databases(model, field,
                                              f"^{re.escape(string)}.*")
        exact_matches = list(filter(lambda x: x.get(field) == string, matches))

        # If multiple matches occur, return the first occurence
        if len(exact_matches) > 1:
            LOGGER.debug("Multiple exact %s matches for %s ! %s", field,
                         model.name.lower(), exact_matches)
            exact_matches = exact_matches[:1]

        # If there are multiple matches and no exact match
        if len(matches) != 1 and len(exact_matches) != 1:
            raise argparse.ArgumentTypeError(
                f"Pattern '{string}' does not identify a single {model.name.lower()}: "
                f"{len(matches)} {model.name.lower()}s match")

        if exact_matches:
            return exact_matches.pop()
        return matches.pop()

    wrapper.__name__ = f"defined_{model.name.lower()}"

    return wrapper


def wildcard_defined_object(model, field):
    """Argument type callback.
    Asserts that the string corresponds to an existing object."""

    def wrapper(string):
        if string == UNSELECTED:
            raise argparse.ArgumentTypeError(
                f"no {model.name} selected nor specified")

        wildcard_string = re.escape(string).replace(r'\#', '.*')
        matches = _search_available_databases(model, field,
                                              f"^{wildcard_string}$")

        if not matches:
            raise argparse.ArgumentTypeError(
                f"Pattern '{string}' does not identify any {model.name.lower()}: "
                f"{len(matches)} {model.name.lower()}s match")
        return matches

    wrapper.__name__ = f"defined_{model.name.lower()}"

    return wrapper



#!/usr/bin/env python3
"""Thin wrapper: generate the static bash completion script for e4s-cl.

Run::

    python scripts/generate_completion_bash.py > completions/e4s-cl.bash
    # or just let main() write the file in-place:
    python scripts/generate_completion_bash.py --in-place

Requires e4s_cl to be importable (``pip install -e .`` or equivalent).
The generated file is committed to the repository and its freshness is
verified in CI.

All generation logic lives in ``e4s_cl/scripts/completion.py``; this script
is just a convenient CLI wrapper.
"""
import sys
from pathlib import Path

# Allow running directly from repo root without a prior install.
sys.path.insert(0, str(Path(__file__).parent.parent))

from e4s_cl.scripts.completion import (  # noqa: E402
    _generate_completion_tree,
    generate_bash,
    regenerate_bash,
)

if __name__ == '__main__':
    if '--in-place' in sys.argv:
        regenerate_bash()
    else:
        print(generate_bash(_generate_completion_tree()), end='')



from pathlib import Path
import tests
from e4s_cl.cf.trace import opened_files


class TraceTest(tests.TestCase):
    def test_opened_files(self):
        returncode, files = opened_files(['cat', '/dev/null'])
        self.assertEqual(returncode, 0)
        self.assertIn(Path('/dev/null'), files)

    def test_return(self):
        returncode, files = opened_files(['cat', '/dev/null'])
        self.assertEqual(returncode, 0)
        for element in files:
            self.assertTrue(isinstance(element, Path))



"""
The **profile detect** command will create a profile from the analysis of \
the given MPI command's execution. This process uses system call \
monitoring to produce an exhaustive list of files and libraries required \
by the MPI installation.

.. warning::
    To get a complete dependency detection, it is best to ensure the following:

    * The MPI launcher and binary should use the host's MPI library, to be \
imported in the future containers;
    * The MPI program should call at least one collective to ensure the use \
of the network stack;
    * The process should be run on multiple nodes using the target network. \
Failure to do so may result in erroneous detection of communication libraries \
and thus may create communication errors when using the profile.

Use :code:`-p/--profile` to select an output profile. If the option is not \
present, the selected profile will be overwritten instead.

.. warning::
   Not specifying a profile will overwrite the selected profile on success !

Examples
----------

.. code::

    mpicc -o ping-pong ping-pong.c
    e4s-cl profile detect -p profile-detected \\
            mpirun -np 2 -ppn 1 -hosts node1,node2 ./ping-pong

"""

import os
import sys
from json import JSONDecodeError
from pathlib import Path
from typing import List, Optional, Any

from sotools import is_elf
from sotools.linker import resolve
from sotools.libraryset import Library, LibrarySet

from e4s_cl import (EXIT_SUCCESS, EXIT_FAILURE, E4S_CL_SCRIPT, logger,
                    INIT_TEMP_PROFILE_NAME)

from e4s_cl import variables
from e4s_cl.error import ProfileSelectionError
from e4s_cl.util import (run_e4scl_subprocess, flatten, json_dumps, json_loads,
                         path_contains, apply_filters, which)
from e4s_cl.cf.trace import opened_files
from e4s_cl.cf.launchers import interpret, get_reserved_directories
from e4s_cl.cli import arguments
from e4s_cl.model.profile import Profile
from e4s_cl.cli.cli_view import AbstractCliView

LOGGER = logger.get_logger(__name__)

LAUNCHER_VAR = '__E4S_CL_DETECT_LAUNCHER'


def _same_file(lhs: Any, rhs: Any) -> bool:
    """Assert two files are the same file, even through symbolic links"""

    def _force_cast(val: Any) -> Path:
        allowed = [str, bytes, os.PathLike]
        for check in allowed:
            if isinstance(val, check):
                return Path(val)
        return Path()

    return _force_cast(lhs).resolve() == _force_cast(rhs).resolve()


def filter_files(path_list: List[Path],
                 launcher: List[str] = None,
                 original_binary: Optional[Library] = None):
    """
    Categorize paths into libraries or files

    Libraries are resolved with the linker and are to be imported in a special
    location. They can only be ELF files.

    Files are referenced using their paths and have to be imported at the same
    location. They can be ELF objects that are dynamically loaded by the library.
    """
    orig_rpath, orig_runpath = [], []

    if original_binary is not None:
        orig_rpath = original_binary.rpath
        orig_runpath = original_binary.runpath

    launcher_reserved_paths = get_reserved_directories(launcher)

    def _not_cache(path: Path) -> bool:
        return path != Path('/etc/ld.so.cache')

    def _not_blacklisted(path: Path) -> bool:
        blacklist = map(Path, ["/tmp", "/sys", "/proc", "/dev", "/run"])
        for entry in blacklist:
            if path_contains(entry, path):
                return False
        return True

    def _not_root(path: Path) -> bool:
        return path != Path('/')

    def _existence(path: Path) -> bool:
        """Assert the file still exists and is accessible"""
        try:
            if not path.exists() or path.is_dir():
                return False
        except PermissionError:
            return False
        return True

    def _waived_launcher(path: Path) -> bool:
        """Assert the file path does not correspond to a directory used by the launcher"""
        for launcher_path in launcher_reserved_paths:
            if path_contains(launcher_path, path):
                return False
        return True

    valid_files = set(filter(_existence, path_list))

    # Apply blacklist and root filters BEFORE is_elf to avoid opening device
    # nodes (e.g. /dev/dri/renderD*) which block on open(), and to skip
    # thousands of /sys/ entries that are never ELF shared libraries.
    valid_files = set(
        apply_filters([_not_blacklisted, _not_root], valid_files))

    elf_objects = set(filter(is_elf, valid_files))
    regular_files = valid_files - elf_objects

    library_set = LibrarySet.create_from(elf_objects)
    if original_binary:
        library_set.add(original_binary)

    def _resolved(path: Path) -> bool:
        """Assert the given path (assuming it to be an elf object) corresponds
        to a library that is resolved via the linker and present in the dynamic
        dependencies of the set.

        This allows us to filter out libraries that are loaded randomly from
        the process from the ones loaded using the standard linux run-time
        linker system. 

        This influences the import method of the objects, as we can control the
        linker system using LD_LIBRARY_PATH, but manually loaded libraries are
        expected to be found at a hardcoded path, and need to be present in
        a container at the same path."""
        # Extract a soname from the given path
        library = Library.from_path(path)

        # Try resolving this soname using the linking rules
        resolved_soname = resolve(library.soname,
                                  rpath=orig_rpath + library_set.rpath,
                                  runpath=orig_runpath + library_set.runpath)

        # For some libraries that disregard SONAME rules (CRAY), try resolving
        # using the file name as it is the one that is actually relevant
        resolved_filename = resolve(path.name,
                                    rpath=orig_rpath + library_set.rpath,
                                    runpath=orig_runpath + library_set.runpath)

        return (_same_file(path, resolved_soname)
                or _same_file(path, resolved_filename)
                or library not in library_set.top_level)

    libraries = set(filter(_resolved, elf_objects))
    orphan_libraries = elf_objects - libraries

    filtered_files = set(
        apply_filters([_not_cache, _waived_launcher],
                      regular_files))
    files = filtered_files.union(orphan_libraries)

    return set(map(str, libraries)), set(map(str, files))


def save_to_profile(profile_name, libraries, files) -> int:
    """
    Save the libraries and files to a profile with the name profile_name
    """
    controller = Profile.controller()
    if profile_name:
        identifier = {'name': profile_name}
        profile = controller.one(identifier)

        if not profile:
            try:
                controller.create(identifier)
            except Exception as err:  # TODO check what errors can arise
                LOGGER.error("Profile creation failed: %s", str(err))
                return EXIT_FAILURE
    else:
        try:
            profile = controller.selected()
        except ProfileSelectionError:
            LOGGER.error("No output profile selected or given as an argument.")
            return EXIT_FAILURE

        if profile.get('name') != INIT_TEMP_PROFILE_NAME:
            LOGGER.warning(
                "No profile specified: currently selected profile will be updated."
            )

        identifier = {'name': profile.get('name')}

    data = {'libraries': list(libraries), 'files': list(files)}
    try:
        controller.update(data, identifier)
    except Exception as err:  # TODO same as above
        LOGGER.error("Profile update failed: %s", str(err))
        return EXIT_FAILURE

    return EXIT_SUCCESS


CORE_MPI_PREFIXES = (
    'libmpi',
    'libmpich',
    'libopen-rte',
    'liborte',
    'liboshmem',
)


def _filter_mpi_artifacts(libs: List[Path], files: List[Path],
                          launcher: Path,
                          mpi_filter: str = 'auto',
                          exclude_prefixes: List[str] = None,
                          exclude_names: List[str] = None) -> (List[str], List[str]):
    """
    Filter MPI-related artifacts using one of three modes:
      - off: no filtering (trust ptrace)
      - manual: apply user exclusions only
      - auto: infer authoritative MPI from libmpi and launcher prefix

    Manual exclusions are always applied before MPI filtering.
    Filtering fails open on ambiguity.
    """
    if not libs or not launcher:
        return [str(p) for p in libs], [str(p) for p in files]

    if mpi_filter == 'off':
        return [str(p) for p in libs], [str(p) for p in files]

    exclude_prefixes = exclude_prefixes or []
    exclude_names = exclude_names or []

    # Normalize exclude prefixes
    normalized_exclude_prefixes = [Path(p).resolve() for p in exclude_prefixes]

    # Warn if exclusions are used in auto mode
    if mpi_filter == 'auto' and (exclude_prefixes or exclude_names):
        LOGGER.warning(
            "Manual exclusions are applied before MPI auto-filtering. "
            "Consider --mpi-filter=manual if this is intentional."
        )

    # Handle manual exclusions first, regardless of auto/manual mode
    if normalized_exclude_prefixes or exclude_names:
        kept_libs = []
        for lib in libs:
            resolved = lib.resolve()
            excluded = False
            if resolved.name in exclude_names:
                excluded = True

            if not excluded:
                for prefix in normalized_exclude_prefixes:
                    if path_contains(prefix, resolved):
                        excluded = True
                        break

            if excluded:
                LOGGER.info("Excluding user-blacklisted artifact %s", lib)
                continue
            kept_libs.append(lib)
        libs = kept_libs

    if mpi_filter == 'manual':
        return [str(p) for p in libs], [str(p) for p in files]

    resolved_launcher = launcher
    if not launcher.is_absolute():
        path_str = which(str(launcher))
        if path_str:
            resolved_launcher = Path(path_str)

    launcher = resolved_launcher.resolve()

    def _get_lib_prefix(path: Path) -> Path:
        """Extract installation prefix from library path.
        
        Looks for 'lib' or 'lib64' directory markers to determine the prefix.
        For example, /opt/mpich/lib/libmpi.so -> /opt/mpich
        Falls back to parent directory for non-standard layouts.
        """
        for marker in ['lib', 'lib64']:
            if marker in path.parts:
                marker_idx = path.parts.index(marker)
                # Ensure marker is a directory component, not the filename
                if marker_idx < len(path.parts) - 1:
                    return Path(*path.parts[:marker_idx])
        # Fallback: use parent directory for non-standard layouts
        # This handles cases like /custom/path/libmpi.so
        return path.parent

    def _is_core_mpi(path: Path) -> bool:
        name = path.name
        return any(name.startswith(prefix) for prefix in CORE_MPI_PREFIXES)

    resolved_libs = [lib.resolve() for lib in libs]

    # 1. Identify authoritative libmpi candidates
    candidates = list(set([
        lib for lib in resolved_libs
        if lib.name == 'libmpi.so' or lib.name.startswith('libmpi.so.')
    ]))

    if not candidates:
        return [str(p) for p in libs], [str(p) for p in files]

    # Check for system launcher
    is_system_launcher = str(launcher).startswith('/usr/bin') or str(launcher).startswith('/bin')

    selected_lib = None

    if is_system_launcher:
        if len(candidates) == 1:
            selected_lib = candidates[0]
        else:
            # Ambiguity or 0 -> fail open
            return [str(p) for p in libs], [str(p) for p in files]
    else:
        # Standard launcher
        if len(candidates) == 1:
            selected_lib = candidates[0]
        else:
            # Multiple candidates. Filter by overlap with launcher prefix.
            launcher_prefix = launcher.parent
            if 'bin' in launcher.parts:
                launcher_prefix = Path(*launcher.parts[:launcher.parts.index('bin')])

            matches = []
            for cand in candidates:
                cand_prefix = _get_lib_prefix(cand)
                # Check overlap
                if (launcher_prefix == cand_prefix or
                        launcher_prefix in cand_prefix.parents or
                        cand_prefix in launcher_prefix.parents):
                    matches.append(cand)

            if len(matches) == 1:
                selected_lib = matches[0]
            else:
                # No match or Multiple matches -> Fail open
                return [str(p) for p in libs], [str(p) for p in files]

    # We have a selected authoritative library
    selected_prefix = _get_lib_prefix(selected_lib)

    LOGGER.info("MPI filtering mode: %s", mpi_filter)
    LOGGER.info("Authoritative MPI: %s", selected_lib)

    kept_libs = []
    discarded_libs = []

    for lib in libs:
        resolved = lib.resolve()

        # Keep non-core libraries always (whitelist)
        if not _is_core_mpi(resolved):
            kept_libs.append(lib)
            continue

        # For core libs, check prefix consistency with authoritative lib
        lib_prefix = _get_lib_prefix(resolved)

        if (selected_prefix == lib_prefix or
                selected_prefix in lib_prefix.parents or
                lib_prefix in selected_prefix.parents):
            kept_libs.append(lib)
        else:
            LOGGER.info("Discarding MPI artifact %s (prefix mismatch with %s)", lib, selected_lib)
            discarded_libs.append(lib)

    if discarded_libs:
        LOGGER.info("Discarded core MPI libs: %s", [l.name for l in discarded_libs])

    return [str(p) for p in kept_libs], [str(p) for p in files]


def detect_subprocesses(launcher, program):
    """Run process profiling in subprocesses with the detected launcher"""
    files, libs = [], []

    os.environ[LAUNCHER_VAR] = launcher[0]
    # If a launcher is present, act as a launcher
    # Use a timeout to prevent hanging on mpirun with multiple processes.
    # The timeout is configurable via __E4S_CL_DETECT_TIMEOUT (seconds).
    # Default is 120s to accommodate srun job-scheduling overhead.
    try:
        detect_timeout = int(os.environ.get('__E4S_CL_DETECT_TIMEOUT', '120'))
    except ValueError:
        detect_timeout = 120
    return_code, json_data = run_e4scl_subprocess([
        *launcher, sys.executable, E4S_CL_SCRIPT, "profile", "detect", *program
    ],
                                                  capture_output=True,
                                                  timeout=detect_timeout)

    if return_code:
        LOGGER.error(
            "Failed to determine necessary libraries: program exited with code %d",
            return_code)
        return libs, files

    # Merge all the data sent back into files and libraries
    file_paths, library_paths = [], []

    for line in json_data.split('\n'):
        try:
            data = json_loads(line)
            file_paths.append(data['files'])
            library_paths.append(data['libraries'])
        except (JSONDecodeError, TypeError):
            pass

    files = list(set(flatten(file_paths)))
    libs = list(set(flatten(library_paths)))

    return libs, files


class ProfileDetectCommand(AbstractCliView):
    """``profile create`` subcommand."""

    def _construct_parser(self):
        usage = f"{self.command} [-p profile] <mpi_launcher command>"
        parser = arguments.get_parser(prog=self.command,
                                      usage=usage,
                                      description=self.summary)

        parser.add_argument('-p',
                            '--profile',
                            help="Output profile",
                            dest='profile_name',
                            metavar='profile_name')

        parser.add_argument('cmd',
                            help="Executable command, e.g. './a.out'",
                            metavar='command',
                            nargs=arguments.REMAINDER)
        
        parser.add_argument('--mpi-filter',
                            help="Filter mode for MPI artifacts (default: auto)",
                            choices=['auto', 'off', 'manual'],
                            default='auto')
        
        parser.add_argument('--exclude-lib-prefix',
                            help="Exclude libraries under this prefix when --mpi-filter=manual",
                            action='append',
                            default=[])

        parser.add_argument('--exclude-lib-name',
                            help="Exclude libraries with this exact name when --mpi-filter=manual",
                            action='append',
                            default=[])

        return parser

    def main(self, argv):
        args = self._parse_args(argv)

        if not args.cmd:
            return EXIT_FAILURE

        launcher, program = interpret(args.cmd)

        if launcher:
            libs, files = detect_subprocesses(launcher, program)

            # Filter out artifacts from other MPI implementations
            libs, files = _filter_mpi_artifacts([Path(l) for l in libs],
                                                [Path(f) for f in files],
                                                Path(launcher[0]),
                                                mpi_filter=getattr(args, 'mpi_filter', 'auto'),
                                                exclude_prefixes=getattr(args, 'exclude_lib_prefix', []),
                                                exclude_names=getattr(args, 'exclude_lib_name', [])
                                                )
        else:
            launcher = os.environ.get(LAUNCHER_VAR, '').split(' ')

            # No launcher, analyse the command
            return_code, accessed_files = opened_files(args.cmd)

            if return_code:
                LOGGER.warning("Command %s failed with error code %d",
                               args.cmd, return_code)

            # Access the binary to check ELF metadata
            binary = None
            if is_elf(args.cmd[0]):
                binary = Library.from_path(args.cmd[0])

            libs, files = filter_files(accessed_files,
                                       launcher,
                                       original_binary=binary)
            LOGGER.debug("Accessed files: %s, %s", libs, files)

        # There are two cases: this is a parent process, in which case we
        # interpret the output of children, or this is a child process, where
        # we just print data on stdout
        if not variables.is_parent():
            print(json_dumps({
                'files': files,
                'libraries': libs,
            }))
            return EXIT_SUCCESS

        return save_to_profile(args.profile_name, libs, files)


COMMAND = ProfileDetectCommand(Profile, __name__)



"""
Profiles are a key component of **e4s-cl**. They are used to greatly \
reduce the amount of information to input on a command line by recording \
chosen arguments.

Instead of listing all necessary files and libraries to use for every command, \
passing a profile containing those files as an argument vastly improves \
readability and efficiency.

Profiles are accessed and edited using the `profile` sub-commands.

Contents
--------

A profile is a recorded collection of fields relating to a specific MPI library.

+------------------+--------------------------------------------------------------+
| Field            | Description                                                  |
+==================+==============================================================+
| `name`           | A name by which the profile is accessed and invoked          |
+------------------+--------------------------------------------------------------+
| `image`          | Path or identifier of a container image                      |
+------------------+--------------------------------------------------------------+
| `backend`        | Identifier for a technology to launch the container with     |
+------------------+--------------------------------------------------------------+
| `source`         | Path of a script to source in the container before execution |
+------------------+--------------------------------------------------------------+
| `files`          | List of files to make accessible to the running program      |
+------------------+--------------------------------------------------------------+
| `libraries`      | List of libraries to load in the running program             |
+------------------+--------------------------------------------------------------+
| `wi4mpi`         | Wi4MPI installation to use for this profile                  |
+------------------+--------------------------------------------------------------+

Profile Creation
----------------

Profiles can be created with the :ref:`init<init>`, \
:ref:`profile detect<profile_detect>` and \
:ref:`profile create<profile_create>` commands.

The commands :ref:`init<init>` and :ref:`profile detect<profile_detect>` will \
create a profile dynamically according to the execution of a reference MPI \
program.

The :ref:`profile create<profile_create>` command will create an empty profile \
to complete at the user's discretion.

.. caution::

    MPI libraries often dynamically use files and libraries without warning, \
and the absence of those unlisted files during execution more often than not \
results in a crash or segmentation fault. Please ensure you acknowledge the \
result of :ref:`profile detect<profile_detect>` when creating an empty profile.

Profile Selection
-----------------

A profile can be selected using the :ref:`profile select<profile_select>` \
command. The target profile is then implicitly used for most of the commands \
taking a profile as an argument.

Only one profile can be selected at a time. Switching selection is done by \
selecting another profile. A selection can also be canceled by using \
:ref:`profile unselect<profile_unselect>`.
"""

from e4s_cl.cli.cli_view import RootCommand
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.detect import COMMAND as detect_command


def _profile_fields():
    for name, data in Profile.__attributes__().items():
        yield " - %(name)s: %(description)s" % {
            'name': name.capitalize(),
            'description': data['description'].capitalize()
        }


HELP_PAGE = """The profile command is used to access and manage profiles.

e4s-cl uses profiles to keep track of execution parameters. These parameters
can then be omitted from the command line given the corresponding profile.

Profiles have the following fields:
%(detail)s

It is highly encouraged to use e4s-cl facilities to create profiles.
The %(detect)s command will complete the selected profile
from the execution of the command passed as an argument. This is the most
secure way to get a accurate profile.

The e4s-cl init command will create a new profile from the available
MPI library, using generic arguments. Those arguments may not correspond to
real use-cases and may result in an incomplete profile.""" % {
    'detect': detect_command.command,
    'detail': "\n".join(_profile_fields())
}

COMMAND = RootCommand(Profile,
                      __name__,
                      group="configuration",
                      help_page_fmt=HELP_PAGE)



"""
Create a new profile as a copy of another existing one.
The first argument is the name of the profile to copy, the second being the destination profile.
"""

from e4s_cl.cli.cli_view import CopyCommand
from e4s_cl.model.profile import Profile

COMMAND = CopyCommand(Profile, __name__)



"""
Detail a profile's content and fields.
The name argument can be omitted in case a profile is selected, in which case the selected profile is shown.
"""

from pathlib import Path
from e4s_cl import EXIT_SUCCESS
from e4s_cl.util import color_text
from e4s_cl.cli.cli_view import ShowCommand
from e4s_cl.model.profile import Profile


def bold(string):
    return color_text(string, 'white', None, ['bold'])


class ProfileShowCommand(ShowCommand):
    """
    Output details about a profile
    """

    def _construct_parser(self):
        parser = super()._construct_parser()

        return parser

    def detail(self, profile_dict):
        """
        Format template for the output
        """

        outline = """%(field_name)s: %(name)s
%(field_image)s: %(image)s
%(field_backend)s: %(backend)s
%(field_script)s: %(script)s
%(field_wi4mpi)s: %(wi4mpi)s

%(field_libs)s:
%(libs)s

%(field_files)s:
%(files)s"""

        headers = {
            'field_name': bold("Profile name"),
            'field_image': bold("Container image"),
            'field_backend': bold("Container tech"),
            'field_script': bold("Pre-execution script"),
            'field_libs': bold("Bound libraries"),
            'field_files': bold("Bound files"),
            'field_wi4mpi': bold("Wi4MPI installation"),
        }

        elements = {
            'name': profile_dict.get('name', 'Not found'),
            'image': profile_dict.get('image', 'Not found'),
            'backend': profile_dict.get('backend', 'Not found'),
            'script': profile_dict.get('source', 'None'),
            'wi4mpi': profile_dict.get('wi4mpi', 'None'),
        }

        if profile_dict.get('libraries'):
            elements['libs'] = "\n".join([
                f" - {p.name} ({p.as_posix()})"
                for p in map(Path, profile_dict['libraries'])
            ])
        else:
            elements['libs'] = "None"

        if profile_dict.get('files'):
            elements['files'] = "\n".join(
                [f" - {p}" for p in profile_dict['files']])
        else:
            elements['files'] = "None"

        print(outline % {**headers, **elements})

    def main(self, argv):
        args = self._parse_args(argv)

        self.detail(args.profile)

        return EXIT_SUCCESS


COMMAND = ProfileShowCommand(Profile, __name__)



import tests
from pathlib import Path
from sotools import Library, linker
from sotools.libraryset import LibrarySet
from e4s_cl.util import which
from e4s_cl.variables import set_dry_run
from e4s_cl.cf.template import Entrypoint
from e4s_cl.cf.containers import Container, FileOptions
import tempfile
from unittest.mock import patch
from e4s_cl.cli.commands.__execute import (import_library, filter_libraries,
                                           overlay_libraries, select_libraries,
                                           COMMAND)


class ExecuteTests(tests.TestCase):

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_lib_import(self):
        container = Container(name="containerless")

        lib = Library.from_path(linker.resolve("libmpi.so"))

        import_library(lib, container)

        links = set()
        for bound in container.bound:
            if bound.origin.resolve() == Path(lib.binary_path).resolve():
                links.add(bound.origin)

        self.assertGreater(len(links), 1)

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_filter_libraries(self):
        lib_set = LibrarySet.create_from(["libmpi.so"])

        self.assertTrue(lib_set.glib)

        filtered = filter_libraries(lib_set, None, None)

        self.assertFalse(filtered.glib)

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_overlay_libraries(self):
        entry = Entrypoint()
        container = Container(name="containerless")
        host_libraries = LibrarySet.create_from([linker.resolve("libmpi.so")])
        host_bash = LibrarySet.create_from([which('bash')])
        lib_set = LibrarySet(host_libraries | host_bash)

        overlain = overlay_libraries(lib_set, container, entry)

        self.assertTrue(lib_set.intersection(overlain))

        self.assertEqual(
            Path(entry.linker).name,
            Path(lib_set.linkers.pop().binary_path).name)

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_select_import_method(self):
        entry = Entrypoint()
        container = Container(name="containerless")
        lib_set = LibrarySet.create_from([linker.resolve("libmpi.so")])

        self.assertTrue(select_libraries(lib_set, container, entry))

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_execute(self):
        set_dry_run(True)

        libmpi = str(linker.resolve('libmpi.so'))

        self.assertCommandReturnValue(0, COMMAND, [
            '--backend',
            'containerless',
            '--image',
            '',
            '--libraries',
            libmpi,
            '--files',
            Path.home().as_posix(),
            'ls',
        ])

        set_dry_run(False)

        self.assertCommandReturnValue(0, COMMAND, [
            '--backend',
            'containerless',
            '--image',
            '',
            '--libraries',
            libmpi,
            'ls',
        ])

        self.assertCommandReturnValue(123, COMMAND, [
            '--backend',
            'containerless',
            '--image',
            '',
            '--libraries',
            libmpi,
            'bash',
            '-c',
            '"exit 123"',
        ])

    @tests.skipIf(not linker.resolve("libmpi.so"), "No test library available")
    def test_execute_alias(self):
        set_dry_run(True)

        libmpi = str(linker.resolve('libmpi.so'))

        with patch('e4s_cl.cf.containers.Container.bind_file') as bind_mock:
            with tempfile.NamedTemporaryFile() as temp_file:
                self.assertCommandReturnValue(0, COMMAND, [
                    '--backend',
                    'containerless',
                    '--image',
                    '',
                    '--libraries',
                    libmpi,
                    '--files',
                    f"{temp_file.name}:/tmp/target",
                    'ls',
                ])

                bind_mock.assert_any_call(temp_file.name, dest='/tmp/target', option=FileOptions.READ_WRITE)

        set_dry_run(False)



from os import getenv, getcwd, environ, pathsep
from tempfile import NamedTemporaryFile, TemporaryDirectory
from unittest import skipIf
from pathlib import Path
import tests
from e4s_cl import config
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
)
from e4s_cl.cf.containers.shifter import _parse_config

SAMPLE_CONFIG = """#system (required)
#
# Name of your system, e.g., edison or cori. This name must match a configured 
# system in the imagegw. This is primarily used by shifterimg to self-identify 
# which system it is representing.
#
system=perlmutter
siteFs=/path1:/path1;\\
    /path2:/path2;
siteEnv=SHIFTER_RUNTIME=1
module_test_siteFs = test
this line is an issue and should be dropped
"""

EXPECTED_CONFIG = dict(system='perlmutter',
                       siteFs='/path1:/path1;/path2:/path2;',
                       siteEnv='SHIFTER_RUNTIME=1',
                       module_test_siteFs='test')

CONFIG_EXECUTABLE = tests.ASSETS / 'bin' / 'shifter-conf'
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
backends:
  shifter:
    executable: '{CONFIG_EXECUTABLE}'
    options: ['--workdir=/opt', '-V', '/opt:/opt:ro']
""")


class ContainerTestShifter(tests.TestCase):

    def test_parse_config(self):
        with NamedTemporaryFile('w', delete=False) as config:
            config.write(SAMPLE_CONFIG)
            config_file = config.name

        directives = _parse_config(config_file)

        self.assertSetEqual(set(EXPECTED_CONFIG.keys()),
                            set(directives.keys()))
        self.assertSetEqual(set(EXPECTED_CONFIG.values()),
                            set(directives.values()))

    def test_create(self):
        container = Container(name='shifter', image='test')
        self.assertFalse(type(container) == Container)
        self.assertTrue(isinstance(container, Container))

    def test_run_image(self):
        container = Container(name='shifter', image='imagenametest')
        command = ['']
        container_cmd = container._prepare(command)
        self.assertIn('imagenametest', ' '.join(map(str, container_cmd)))

    def test_run_mpirun(self):
        container = Container(name='shifter', image='dummyimagename')
        command = ['mpirun -n 2 ls']
        container_cmd = container._prepare(command)
        self.assertIn(command[0], ' '.join(map(str, container_cmd)))

    def test_bind_file(self):
        container = Container(name='shifter')

        target = Path('/tmp')
        dest = Path('/etc')
        contained_dest = Path("/etc/skel")
        option = FileOptions.READ_WRITE

        container.bind_file(target)
        self.assertIn(BoundFile(target, target, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=contained_dest)
        self.assertIn(BoundFile(target, contained_dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

    def test_bind_relative(self):
        container = Container(name='shifter')

        target = Path('/tmp/../proc/meminfo')

        ref = Path('/tmp')
        file = Path('/proc/meminfo')

        container.bind_file(target)
        files = set(map(lambda x: x.origin, container.bound))

        self.assertSetEqual({ref, file}, files)

    def test_prepare_import_container_dir(self):
        """
        Assert CONTAINER_DIR imports will trigger the creation of a mock directory
        """

        def contents(directory: Path):
            return set(map(lambda x: x.name, directory.iterdir()))

        container = Container(name='shifter')
        temp = TemporaryDirectory()
        path = Path(temp.name)

        container.bind_file(
            Path(__file__),
            Path(container.import_binary_dir) / Path(__file__).name)

        volumes = container._setup_import(path)

        self.assertIn(container.import_binary_dir.name, contents(path))
        self.assertIn(
            Path(__file__).name,
            contents(path / container.import_binary_dir.name))

        temp.cleanup()

    def test_prepare_import_etc_files(self):
        """
        Assert importing /etc files fails
        """

        def contents(directory: Path):
            return set(map(lambda x: x.name, directory.iterdir()))

        container = Container(name='shifter')
        temp = TemporaryDirectory()
        path = Path(temp.name)

        # Files to attempt binding, must ideally be present on every system
        files = ['/etc/ld.so.cache', '/etc/ld.so.conf.d/']

        for file in files:
            container.bind_file(file, file)

        volumes = container._setup_import(path)

        for bind in volumes:
            for file in files:
                self.assertNotIn(file, volumes)

        temp.cleanup()

    def test_additional_options_config(self):
        container = Container(name='shifter')
        command = ['']

        shifter_command = container._prepare(command)
        for option in {'--workdir=/opt', '-V', '/opt:/opt:ro'}:
            self.assertNotIn(option, shifter_command)

        config.update_configuration(TEST_CONFIGURATION)
        shifter_command = container._prepare(command)
        self.assertContainsInOrder([
            '--workdir=/opt',
            '-V',
            '/opt:/opt:ro',
        ], shifter_command)

        config.update_configuration(DEFAULT_CONFIGURATION)
        shifter_command = container._prepare(command)
        for option in {'--workdir=/opt', '-V', '/opt:/opt:ro'}:
            self.assertNotIn(option, shifter_command)

    def test_additional_options_environment(self):
        container = Container(name='shifter')
        command = ['']

        shifter_command = container._prepare(command)
        for option in {'--workdir=/opt', '-V', '/opt:/opt:ro'}:
            self.assertNotIn(option, shifter_command)

        environ['E4S_CL_SHIFTER_OPTIONS'] = "--workdir=/opt -V /opt:/opt:ro"
        shifter_command = container._prepare(command)
        self.assertContainsInOrder([
            '--workdir=/opt',
            '-V',
            '/opt:/opt:ro',
        ], shifter_command)

        del environ['E4S_CL_SHIFTER_OPTIONS']
        shifter_command = container._prepare(command)
        for option in {'--workdir=/opt', '-V', '/opt:/opt:ro'}:
            self.assertNotIn(option, shifter_command)

    def test_executable(self):
        """Assert the default shifter executable comes from $PATH"""
        container = Container(name='shifter')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"

        self.assertEqual(tests.ASSETS / 'bin' / 'shifter',
                         container._executable())

        environ['PATH'] = default_path

    def test_executable_config(self):
        """Assert the shifter executable is read from the configuration"""
        container = Container(name='shifter')

        config.update_configuration(TEST_CONFIGURATION)
        self.assertEqual(tests.ASSETS / 'bin' / 'shifter-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

    def test_executable_env(self):
        """Assert the shifter executable is read from the environment"""
        container = Container(name='shifter')

        environ['E4S_CL_SHIFTER_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                   'shifter-env')
        self.assertEqual(tests.ASSETS / 'bin' / 'shifter-env',
                         container._executable())

        del environ['E4S_CL_SHIFTER_EXECUTABLE']

    def test_executable_priority(self):
        """Assert the environment has precedence over config and config over default"""
        container = Container(name='shifter')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"
        config.update_configuration(TEST_CONFIGURATION)
        environ['E4S_CL_SHIFTER_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                   'shifter-env')

        self.assertEqual(tests.ASSETS / 'bin' / 'shifter-env',
                         container._executable())

        del environ['E4S_CL_SHIFTER_EXECUTABLE']

        self.assertEqual(tests.ASSETS / 'bin' / 'shifter-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

        self.assertEqual(tests.ASSETS / 'bin' / 'shifter',
                         container._executable())

        environ['PATH'] = default_path



"""
This command is a helper meant to easily show differences between profiles.
It will output a list of paths prefixed by a inequality sign \
        (:code:`>`/:code:`<`) to highlight which profile has an outstanding \
        dependency.

Example
--------

.. code::

    $ e4s-cl profile diff openmpi sparse
    < /usr/lib/openmpi/libopen-pal.so.40  # Files only in the 'openmpi' profile
    < /usr/lib/openmpi/libopen-rte.so.40
    > /usr/bin/strace                     # File only in the 'sparse' module

"""

from e4s_cl.cli import arguments
from e4s_cl.cli.cli_view import AbstractCliView
from e4s_cl.model.profile import Profile


class DiffCommand(AbstractCliView):
    """
    Command outlining differences between models
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('summary_fmt', "Compare %(model_name)ss.")
        super().__init__(*args, **kwargs)

    def _construct_parser(self):
        key_attr = self.model.key_attribute
        parser = arguments.get_model_identifier(self.model,
                                                prog=self.command,
                                                description=self.summary)

        parser.add_argument(self.model_name + '_rhs',
                            type=arguments.single_defined_object(
                                self.model, key_attr),
                            help="The profile to compare with",
                            metavar=f"other_{self.model_name}_{key_attr}")

        return parser

    def main(self, argv):
        args = self._parse_args(argv)

        lhs = getattr(args, self.model_name)
        rhs = getattr(args, self.model_name + '_rhs')

        if not (lhs and rhs):
            self.parser.error("Missing profile argument")

        def _order_r(lhs, rhs):
            return (list(set(lhs) - set(rhs)), '<')

        def _order_l(lhs, rhs):
            return (list(set(rhs) - set(lhs)), '>')

        def _diff_member(attr, order):
            diff, sign = order(lhs.get(attr, []), rhs.get(attr, []))
            diff.sort()

            for element in diff:
                print(f"{sign} {element}")

        _diff_member('libraries', _order_r)
        _diff_member('files', _order_r)
        _diff_member('libraries', _order_l)
        _diff_member('files', _order_l)

        return 0


COMMAND = DiffCommand(Profile, __name__)



import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from e4s_cl.cli.commands.profile.detect import _filter_mpi_artifacts

class FilterMPITest(unittest.TestCase):
    def setUp(self):
        self.tmp_dir_obj = TemporaryDirectory()
        self.tmp_dir = Path(self.tmp_dir_obj.name)
        
        # Create 'system' MPI
        self.sys_mpi = self.tmp_dir / "usr/lib/mpich/lib/libmpi.so"
        self.sys_mpi.parent.mkdir(parents=True)
        self.sys_mpi.touch()

        # Create 'system' launcher
        self.sys_launcher = self.tmp_dir / "usr/bin/mpirun"
        self.sys_launcher.parent.mkdir(parents=True)
        self.sys_launcher.touch()
        self.sys_launcher.chmod(0o755)

        # Create 'conflicting' Intel MPI
        self.intel_mpi = self.tmp_dir / "opt/intel/oneapi/lib/libmpi.so"
        self.intel_mpi.parent.mkdir(parents=True)
        self.intel_mpi.touch()

        # Create unrelated artifact (depends on Intel MPI)
        self.intel_dep = self.tmp_dir / "opt/intel/oneapi/lib/libfabric.so"
        self.intel_dep.touch()

    def tearDown(self):
        self.tmp_dir_obj.cleanup()

    def test_filter_pollution(self):
        libs = [self.sys_mpi, self.intel_mpi, self.intel_dep]
        files = []
        
        filtered_libs, filtered_files = _filter_mpi_artifacts(libs, files, self.sys_launcher)
        
        # We expect sys_mpi to be present
        self.assertIn(str(self.sys_mpi), filtered_libs)
        
        # We expect intel_mpi to be removed
        self.assertNotIn(str(self.intel_mpi), filtered_libs)
        
        # We expect intel_dep to be KEPT now, because we whitelist non-core libs
        self.assertIn(str(self.intel_dep), filtered_libs)

    def test_filter_system_usr(self):
        # Scenario: Launcher in /usr/bin, lib in /usr/lib
        # This checks the "system" heuristic
        
        libs = [self.sys_mpi]
        files = []
        filtered_libs, _ = _filter_mpi_artifacts(libs, files, self.sys_launcher)
        self.assertIn(str(self.sys_mpi), filtered_libs)

    def test_filter_divergent_paths_kept(self):
        """
        Divergent paths (where launcher and lib don't share prefix) are kept 
        if no other library matches the launcher. Fail-open behavior.
        """
        divergent_dir = self.tmp_dir / "opt/divergent"
        divergent_dir.mkdir(parents=True)
        
        launcher = divergent_dir / "bin/mpirun"
        launcher.parent.mkdir(parents=True)
        launcher.touch()
        
        lib = divergent_dir / "custom/lib/libmpi.so"
        lib.parent.mkdir(parents=True)
        lib.touch()
        
        libs = [lib]
        files = []
        filtered_libs, _ = _filter_mpi_artifacts(libs, files, launcher)
        
        self.assertIn(str(lib), filtered_libs)

    def test_example_a_openmpi_external_deps(self):
        """
        Example A: Open MPI built with external PMIx / UCX / HWLOC
        Should keep all libraries because they are required deps, even if different prefixes.
        """
        launcher = Path('/opt/mpi/openmpi/4.1.6/bin/mpirun')
        libs = [
            Path('/opt/mpi/openmpi/4.1.6/lib/libmpi.so'),
            Path('/opt/pmix/4.2.6/lib/libpmix.so'),
            Path('/opt/ucx/1.15.0/lib/libucp.so'),
            Path('/usr/lib64/libhwloc.so')
        ]
        files = []
        
        kept_libs, kept_files = _filter_mpi_artifacts(libs, files, launcher)
        
        self.assertEqual(len(kept_libs), 4)
        self.assertIn('/opt/mpi/openmpi/4.1.6/lib/libmpi.so', kept_libs)

    def test_example_c_vendor_wrapper(self):
        """
        Example C: Vendor launcher + Spack-provided MPI
        Launcher in /usr/bin, MPI in /opt/spack via LD_LIBRARY_PATH.
        Should keep everything.
        """
        launcher = Path('/usr/bin/mpirun')
        libs = [
            Path('/opt/spack/openmpi/lib/libmpi.so'),
            Path('/opt/spack/pmix/lib/libpmix.so')
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(libs, files, launcher)

        self.assertEqual(len(kept_libs), 2)

    def test_example_e_worst_case(self):
        """
        Example E: Mixed environment with /usr/bin launcher
        """
        launcher = Path('/usr/bin/mpirun')
        libs = [
            Path('/opt/spack/openmpi/lib/libmpi.so'),
            Path('/opt/spack/pmix/lib/libpmix.so'),
            Path('/usr/lib64/libhwloc.so'),
            Path('/opt/intel/oneapi/lib/libfabric.so')
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(libs, files, launcher)

        self.assertEqual(len(kept_libs), 4)

    def test_multiple_mpi_core_libs(self):
        """
        Scenario where multiple MPI core libs exist.
        Should pick authoritative one (closest to launcher) and discard other CORE libs.
        But keep non-core libs even if under discarded prefix.
        """
        launcher = Path('/opt/ompi/bin/mpirun')
        libs = [
            Path('/opt/ompi/lib/libmpi.so'),        # Authoritative
            Path('/opt/mpich/lib/libmpich.so'),     # Conflicting MPI
            Path('/opt/mpich/lib/libfabric.so'),    # Useful dep in discarded prefix
            Path('/opt/other/lib/libfoo.so')        # Neutral
        ]
        
        kept_libs, kept_files = _filter_mpi_artifacts(libs, files=[], launcher=launcher)
        
        self.assertIn('/opt/ompi/lib/libmpi.so', kept_libs)
        self.assertNotIn('/opt/mpich/lib/libmpich.so', kept_libs)
        self.assertIn('/opt/mpich/lib/libfabric.so', kept_libs)
        self.assertIn('/opt/other/lib/libfoo.so', kept_libs)

    def test_mpi_filter_off(self):
        """
        Test disabling the filter.
        """
        launcher = Path('/opt/ompi/bin/mpirun')
        libs = [
            Path('/opt/ompi/lib/libmpi.so'),
            Path('/opt/mpich/lib/libmpich.so'),
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(libs, files, launcher, mpi_filter='off')
        
        self.assertEqual(len(kept_libs), 2)

    def test_mpi_filter_manual(self):
        """
        Test manual filter (disables auto filter, but respects exclusions if provided - tested separately).
        Solely manual should return everything if no exclusions provided.
        """
        launcher = Path('/opt/ompi/bin/mpirun')
        libs = [
            Path('/opt/ompi/lib/libmpi.so'),
            Path('/opt/mpich/lib/libmpich.so'),
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(libs, files, launcher, mpi_filter='manual')

        self.assertEqual(len(kept_libs), 2)

    def test_exclude_lib_name(self):
        """
        Test manual exclusion by name.
        """
        launcher = Path('/opt/ompi/bin/mpirun')
        libs = [
            Path('/opt/ompi/lib/libmpi.so'),
            Path('/opt/mpich/lib/libmpich.so'),
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(
            libs, files, launcher, 
            mpi_filter='auto', 
            exclude_names=['libmpich.so']
        )
        
        self.assertIn('/opt/ompi/lib/libmpi.so', kept_libs)
        self.assertNotIn('/opt/mpich/lib/libmpich.so', kept_libs)

    def test_exclude_lib_prefix(self):
        """
        Test manual exclusion by prefix.
        """
        launcher = Path('/opt/ompi/bin/mpirun')
        libs = [
            Path('/opt/ompi/lib/libmpi.so'),
            Path('/opt/bad/lib/libbad.so'),
        ]
        files = []

        kept_libs, kept_files = _filter_mpi_artifacts(
            libs, files, launcher, 
            mpi_filter='auto', 
            exclude_prefixes=['/opt/bad']
        )
        
        self.assertIn('/opt/ompi/lib/libmpi.so', kept_libs)
        self.assertNotIn('/opt/bad/lib/libbad.so', kept_libs)



import pathlib
import tests
from e4s_cl.cli.commands.__main__ import COMMAND as main


class VersionTest(tests.TestCase):
    def test_output(self):
        argv = ['-V']
        stdout, stderr = self.assertCommandReturnValue(0, main, argv)



"""launchers module
Entry point for supported program launchers"""

import re
import sys
from os import environ
from types import ModuleType
from typing import (
    Dict,
    List,
    Tuple,
    Optional,
)
from pathlib import Path
from importlib import import_module
from shlex import split
from e4s_cl import logger, config
from e4s_cl.util import (
    get_env,
    walk_packages,
)
from e4s_cl.error import InternalError

LOGGER = logger.get_logger(__name__)


class Parser:
    """
    Default parser.

    Uses a dictionary of known options when available, but applies robust
    fallbacks so unknown options don't prematurely terminate parsing.

    Rules:
    - '--' ends option parsing; remainder is the program.
    - ':' (Open MPI app-context delimiter) is retained on the launcher side.
      Note: multi-app context still needs higher-level handling to inject the
      wrapper before each app.
    - Known options consume their declared nargs.
    - '--opt=value' is always accepted.
    - Unknown long options ('--opt') consume one value if the next token isn't
      an option; otherwise consume none.
    - Unknown short options consume one value if the next token isn't an option.
      For concatenated known 1-arg short opts (e.g., '-xFOO=bar'), the entire
      token is accepted as one.
    """

    def __init__(self, arguments: Dict[str, int]):
        self.arguments = arguments or {}

    @staticmethod
    def _is_option(tok: str) -> bool:
        return tok.startswith('-') and tok != '-'

    @staticmethod
    def _has_inline_value(tok: str) -> bool:
        return tok.startswith('--') and '=' in tok

    def parse(self, command: List[str]) -> Tuple[List[str], List[str]]:
        """
        Separate a command line into launcher and program.
        """
        if not command:
            return [], []

        position = 0
        launcher = [command[position]]
        position += 1
        n = len(command)

        while position < n:
            flag = command[position]

            # End-of-options sentinel: keep it on the launcher side and stop
            if flag == '--':
                launcher.append(flag)
                position += 1
                break

            # Open MPI app-context delimiter: retain and continue parsing
            if flag == ':':
                launcher.append(flag)
                position += 1
                continue

            # Known option: consume declared nargs
            if flag in self.arguments:
                nargs = self.arguments[flag]
                launcher.append(flag)
                # Append up to nargs following tokens safely
                for k in range(min(nargs, n - position - 1)):
                    launcher.append(command[position + 1 + k])
                position += 1 + nargs
                continue

            # Generic '--opt=value'
            if self._has_inline_value(flag):
                launcher.append(flag)
                position += 1
                continue

            # Unknown long option: '--opt'
            if flag.startswith('--'):
                launcher.append(flag)
                # Heuristic: if next token is not an option, treat as value
                if position + 1 < n and not self._is_option(command[position + 1]):
                    launcher.append(command[position + 1])
                    position += 2
                else:
                    position += 1
                continue

            # Short options (unknown or concatenated)
            if flag.startswith('-') and flag != '-':
                # If any known 1-arg option is a prefix of this token, accept concatenated form
                concatenated = next(
                    (
                        opt for opt, nargs in self.arguments.items()
                        if nargs == 1 and flag.startswith(opt)
                    ),
                    None,
                )
                if concatenated:
                    launcher.append(flag)
                    position += 1
                    continue

                # Otherwise, assume unknown short option possibly with one value
                launcher.append(flag)
                if position + 1 < n and not self._is_option(command[position + 1]):
                    launcher.append(command[position + 1])
                    position += 2
                else:
                    position += 1
                continue

            # First non-option token: program starts here
            break

        return launcher, command[position:]

LAUNCHERS = {}

for _, _module_name, _ in walk_packages(path=__path__, prefix=__name__ + '.'):
    import_module(name=_module_name)
    for script_name in sys.modules[_module_name].SCRIPT_NAMES:
        LAUNCHERS.update({script_name: _module_name})


def get_launcher(cmd: List[str]) -> Optional[ModuleType]:
    """
    Return a launcher module for the given command
    """
    if not cmd or cmd is None:
        return None

    script = Path(cmd[0]).name
    module_name = LAUNCHERS.get(script)

    if not module_name:
        return None
    return sys.modules[module_name]


def get_reserved_directories(cmd: List[str]) -> List[Path]:
    launcher_module = get_launcher(cmd)
    if launcher_module is not None and getattr(launcher_module, 'META', False):
        if 'reserved_directories' in launcher_module.META:
            return list(map(Path,
                            launcher_module.META['reserved_directories']))

    return []


def _additional_options() -> List[str]:
    marker = "launcher_options"

    env_options = get_env(marker)
    if env_options:
        return split(env_options)

    config_options = getattr(
        config.CONFIGURATION,
        marker,
        None,
    )

    if config_options:
        return config_options

    return []


def interpret(cmd: List[str]) -> Tuple[List[str], List[str]]:
    """
    Tries its best to understand what part of a command line is the launcher
    and what part is the application
    """
    # Assert cmd[0] is valid by terminating for empty lists
    if not cmd:
        return [], []

    def _parse_from_mod(cmd):
        module = get_launcher(cmd)

        if module is None:
            raise NotImplementedError(
                f"Launcher {Path(cmd[0]).name} is not supported")

        return module.PARSER.parse(cmd)

    # If '--' appears in the command then everything before it is a launcher + args
    # and everything after is the application + args
    if '--' in cmd:
        idx = cmd.index('--')
        launcher, application = cmd[:idx], cmd[idx + 1:]
    elif Path(cmd[0]).name in LAUNCHERS:
        launcher, application = _parse_from_mod(cmd)
    else:
        return [], cmd

    return [*launcher, *_additional_options()], application


def filter_arguments(parser: Parser,
                     cmd: List[str]) -> Tuple[List[str], List[str]]:
    """Filter a list of arguments into the ones defined by the given parser
    and foreign ones, and return them in two lists"""
    valid, foreign = [], []

    tokens = list(cmd)
    while tokens:
        token = tokens.pop(0)
        if token in parser.arguments:
            to_consume = min(parser.arguments[token], len(tokens))
            valid.append(token)
            for _ in range(to_consume):
                valid.append(tokens.pop(0))
        else:
            foreign.append(token)

    return valid, foreign



#pylint: disable=E1131
"""
Load and propagate the contents of configuration files in YAML format
"""

import sys
import yaml
from dataclasses import dataclass
from typing import Dict
from pathlib import Path
from e4s_cl import (
    CONTAINER_DIR,
    E4S_CL_HOME,
    E4S_CL_TEST,
    EXIT_FAILURE,
    PROFILE_LIST_DEFAULT_COLUMNS,
    WI4MPI_DEFAULT_INSTALL_DIR,
)


def update_configuration(configuration):
    global CONFIGURATION
    CONFIGURATION = configuration


def flatten(data):
    """
    Transform nested dictionaries into key value pairs by prefixing the
    parent's key, under the assumption that all keys are str.
    >>> flatten({'root': {'key1': 0, 'key2': 'test'} })

    => dict(root_key1=0, root_key2='test')
    """
    separator = '_'

    def _pre(prefix, string):
        return str(separator.join(filter(None, [prefix, string])))

    def _intermediate(prefix, data: dict):
        flat = {}

        if not data:
            return flat

        for key, value in data.items():
            if isinstance(value, dict):
                for ckey, cval in _intermediate(str(key), value).items():
                    flat.update({_pre(prefix, ckey): cval})
            else:
                flat.update({_pre(prefix, key): value})

        return flat

    return _intermediate('', data)


@dataclass(frozen=True)
class ConfigurationField:
    key: str
    expected_type: type
    default: callable
    description: str = ""


@dataclass(frozen=True)
class ConfigurationGroup:
    key: str
    fields: frozenset  # [ConfigurationField|ConfigurationGroup]
    description: str = ""

    def __init__(self, key, fields, description=""):
        object.__setattr__(self, "key", key)
        object.__setattr__(self, "fields", frozenset(fields))
        object.__setattr__(self, "description", description)

    def flatten(self):
        _fields = []

        for field in self.fields:
            if isinstance(field, ConfigurationGroup):
                _fields.extend(field.flatten())
            elif isinstance(field, ConfigurationField):
                _fields.append(field)

        for field in _fields:
            namespaced = "_".join(filter(None, [self.key, field.key]))
            yield ConfigurationField(namespaced, field.expected_type,
                                     field.default)

    def as_dict(self) -> Dict:
        out = {}

        for field in self.fields:
            if isinstance(field, ConfigurationGroup):
                out[field.key] = field.as_dict()

            elif isinstance(field, ConfigurationField):
                out[field.key] = field.default()

        return out

    def template(self):
        return yaml.safe_dump(self.as_dict())


class ConfigurationError(Exception):
    """
    Do not pass through logger nor error as those depend on configuration
    """

    def handle(self, _etype, _value, _tb):
        print(self.args[0], file=sys.stderr)
        return EXIT_FAILURE


class Configuration:
    """
    Class of objects abstracting configuration values. Can be created using a
    dict or class methods to complete it with defined configuration fields,
    then merged with other Configuration objects using the bitwise or operation.
    """

    @classmethod
    def create_from_string(cls, string, complete=False):
        """
        Create a Configuration object from a YAML string, with type checking.
        Will complete with default values for missing fields if complete is set
        to True.
        """

        config = cls()
        data = flatten(yaml.safe_load(string)) or {}

        for parameter in ALLOWED_CONFIG.flatten():
            field = {}
            if parameter.key in data:
                value = data[parameter.key]

                if isinstance(value, parameter.expected_type):
                    field = {parameter.key: value}
                else:
                    raise ConfigurationError(
                        f"Invalid value for parameter '{parameter.key}':"
                        f"{value} (expected {str(parameter.expected_type)})")

            elif complete:
                field = {parameter.key: parameter.default()}

            config._fields.update(field)

        return config

    @classmethod
    def create_from_file(cls, config_file, complete=False):
        yaml_contents = ''
        if config_file and Path(config_file).exists():
            with open(config_file, encoding='utf-8') as file:
                yaml_contents = file.read()

        return Configuration.create_from_string(yaml_contents,
                                                complete=complete)

    @classmethod
    def default(cls):
        return cls.create_from_string('', complete=True)

    def __init__(self, defaults=None):
        if isinstance(defaults, dict):
            self._fields = defaults
        else:
            self._fields = {}

    def __getattr__(self, identifier):
        if identifier in self._fields:
            return self._fields[identifier]
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{identifier}'"
        )

    def __or__(self, rhs):
        """
        Merge configuration objects
        The right hand side has priority, and its keys will take precedence
        """
        if not isinstance(rhs, Configuration):
            raise TypeError(f"unsupported operand type(s) for |: "
                            f"'{type(self)}' and '{type(rhs)}'")

        # Merge the two dictionaries (Use | on py39)
        return Configuration({**self._fields, **rhs._fields})

    def __str__(self):
        return str(self._fields)


USER_CONFIG_PATH = Path.home() / ".config/e4s-cl.yaml"
INSTALL_CONFIG_PATH = Path(E4S_CL_HOME) / "e4s-cl.yaml"
SYSTEM_CONFIG_PATH = "/etc/e4s-cl/e4s-cl.yaml"

ALLOWED_CONFIG = ConfigurationGroup(
    "", {
        ConfigurationField(
            "container_directory",
            str,
            lambda: CONTAINER_DIR,
            "e4s-cl data directory location inside the container",
        ),
        ConfigurationField(
            "launcher_options",
            list,
            lambda: [],
            "Additional options to pass to the process launcher",
        ),
        ConfigurationField(
            "profile_list_columns",
            list,
            lambda: PROFILE_LIST_DEFAULT_COLUMNS,
            "Columns to display with the profile list command",
        ),
        ConfigurationField(
            "preload_root_libraries",
            bool,
            lambda: True,
            "Insert LD_PRELOAD calls to ensure bound libraries are preloaded. Required when using RPATH'ed libraries",
        ),
        ConfigurationField(
            "disable_ranked_log",
            bool,
            lambda: False,
            "Disable logging on the work nodes",
        ),
        ConfigurationGroup(
            "wi4mpi",
            {
                ConfigurationField(
                    "install_directory",
                    str,
                    lambda: str(WI4MPI_DEFAULT_INSTALL_DIR),
                    "Location of the Wi4MPI installation to use. If the directory does not exist, e4s-cl will attempt to install Wi4MPI in that directory.",
                )
            },
            "Wi4MPI-related configuration options",
        ),
        ConfigurationGroup(
            "backends", {
                ConfigurationGroup(
                    "singularity",
                    {
                        ConfigurationField(
                            "executable",
                            str,
                            lambda: "",
                            "Location of the singularity executable to use",
                        ),
                        ConfigurationField(
                            "options",
                            list,
                            lambda: [],
                            "Options to pass to the singularity executable",
                        ),
                        ConfigurationField(
                            "exec_options",
                            list,
                            lambda: [],
                            "Options to pass to the singularity exec command",
                        ),
                    },
                    "Singularity container backend configuration",
                ),
                ConfigurationGroup(
                    "apptainer",
                    {
                        ConfigurationField(
                            "executable",
                            str,
                            lambda: "",
                            "Location of the apptainer executable to use",
                        ),
                        ConfigurationField(
                            "options",
                            list,
                            lambda: [],
                            "Options to pass to the apptainer executable",
                        ),
                        ConfigurationField(
                            "exec_options",
                            list,
                            lambda: [],
                            "Options to pass to the apptainer exec command",
                        ),
                    },
                    "Apptainer container backend configuration",
                ),
                ConfigurationGroup(
                    "podman", {
                        ConfigurationField(
                            "executable",
                            str,
                            lambda: "",
                            "Location of the podman executable to use",
                        ),
                        ConfigurationField(
                            "options",
                            list,
                            lambda: [],
                            "Options to pass to the podman executable",
                        ),
                        ConfigurationField(
                            "run_options",
                            list,
                            lambda: [],
                            "Options to pass to the podman run command",
                        ),
                    }),
                ConfigurationGroup(
                    "shifter", {
                        ConfigurationField(
                            "executable",
                            str,
                            lambda: "",
                            "Location of the shifter executable to use",
                        ),
                        ConfigurationField(
                            "options",
                            list,
                            lambda: [],
                            "Options to pass to the shifter executable",
                        ),
                    }),
                ConfigurationGroup(
                    "barebones",
                    {
                        ConfigurationField(
                            "options",
                            list,
                            lambda: [],
                            "Options to pass before the execution script",
                        ),
                        ConfigurationField(
                            "exec_options",
                            list,
                            lambda: [],
                            "Options to pass after the execution script",
                        ),
                    },
                    "Barebones backend configuration",
                ),
            }),
    })

CONFIGURATION = Configuration.default()
if not E4S_CL_TEST:
    CONFIGURATION = CONFIGURATION  \
        | Configuration.create_from_file(SYSTEM_CONFIG_PATH) \
        | Configuration.create_from_file(INSTALL_CONFIG_PATH) \
        | Configuration.create_from_file(USER_CONFIG_PATH)



"""
Error-path tests for the 'init' command (e4s_cl/cli/commands/init.py).

These tests exercise argument-validation failures that occur before any MPI
execution or tracing takes place.  They are therefore portable and require no
MPI installation or running container daemon.
"""

import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.init import COMMAND


class InitArgumentErrorTest(tests.TestCase):
    """Argument validation errors in InitCommand.main()."""

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_bad_launcher_fails(self):
        """--launcher with a binary not on PATH must fail at argument parsing time."""
        self.assertNotCommandReturnValue(
            0, COMMAND, ['--launcher', 'no_such_launcher_4e2scl_test'])

    def test_bad_launcher_with_extra_options_fails(self):
        """Combining a bad --launcher with other valid options still fails."""
        self.assertNotCommandReturnValue(
            0, COMMAND,
            ['--launcher', 'no_such_launcher_4e2scl_test',
             '--launcher_args', '-n 2',
             '--profile', 'my_profile'])



"""
Module introducing singularity support
"""

import os
from pathlib import Path
from typing import List, Union, Optional
from e4s_cl import logger, BAREBONES_SCRIPT, BAREBONES_LIBRARY_DIR
from e4s_cl.util import run_subprocess, create_symlink, empty_dir, mkdirp, list_directory_files
from e4s_cl.cf.libraries import cache_libraries
from e4s_cl.cf.containers import Container, FileOptions
from e4s_cl.cf.wi4mpi import wi4mpi_root

LOGGER = logger.get_logger(__name__)

NAME = 'barebones'
MIMES = []

OPTION_STRINGS = {FileOptions.READ_ONLY: 'ro', FileOptions.READ_WRITE: 'rw'}


class BarebonesContainer(Container):
    """
    Class to use when formatting bound files for a singularity execution
    """

    executable_name = ''

    def __init__(self, *args, **kwargs):
        if Path(BAREBONES_LIBRARY_DIR).is_dir():
            empty_dir(Path(BAREBONES_LIBRARY_DIR))
        else:
            mkdirp(Path(BAREBONES_LIBRARY_DIR))
        super().__init__(*args, **kwargs)

    def _working_dir(self):
        return ['--pwd', os.getcwd()]

    def __setup__(self):
        """
        Bind minimal directories stripped from the use of `--contain`
        """
        # Ensure HOME is bound
        #self.bind_file(Path.home(), option=FileOptions.READ_WRITE)

        # To use with --contain, but removed as it prevented PMI setup on Theta
        #self.bind_file('/dev', option=FileOptions.READ_WRITE)
        #self.bind_file('/tmp', option=FileOptions.READ_WRITE)

    @property
    def script(self):
        return Path(BAREBONES_SCRIPT)

    @property
    def import_library_dir(self):
        return Path(BAREBONES_LIBRARY_DIR)

    def _format_bound(self):
        """
        Format a list of files to a compatible bind option of singularity
        """

        def _format():
            for file in self.bound:
                yield f"{file.origin}:{file.destination}:{OPTION_STRINGS[file.option]}"

        self.env.update({"BAREBONES_BIND": ','.join(_format())})


    def list_directory_sofiles(self, path: Path):
        """Lists all the so files of a directory.

        Args:
            path (Path): path of the directory list the so files of.

        Returns:
            A list of paths of the so files in the given directory.
        """
        file_paths = list_directory_files(path)
        sofile_paths = []
        for file_path in file_paths:
            if '.so' in file_path.suffixes: # Check if it is a library file
                sofile_paths.append(file_path.absolute())
        return sofile_paths

    def _prepare(self, command: List[str], overload: bool = True) -> List[str]:
        """
        Return the command to run in a list of string
        """

        # Chech the environment for the use of Wi4MPI
        wi4mpi_install_dir = wi4mpi_root()
        # If WI4MPI is to be used, we don't preload the mpi's libraries
        if wi4mpi_install_dir is None:
            to_be_preloaded = self.list_directory_sofiles(Path(BAREBONES_LIBRARY_DIR))
            for file_path in to_be_preloaded:
                self.add_ld_preload(str(file_path))
            self.env.update(
                {'LD_PRELOAD': ":".join(self.ld_preload)})

        # LD_LIBRARY_PATH override does not respect container's values.
        # Enabling this may prevent crashes with nvidia library import
        # from singularity (--nv flag) but causes MORE crashes with client containers
        # self.env.update({'SINGULARITYENV_LD_LIBRARY_PATH': ":".join(self.ld_lib_path)})
        self._format_bound()
        nvidia_flag = ['--nv'] if self._has_nvidia() else []

        return [
            *self._additional_options(),
            *command,
            *self._additional_options('exec'),
        ]

    def bind_file(self,
                  path: Union[Path, str],
                  dest: Optional[Path] = None,
                  option: int = FileOptions.READ_ONLY) -> None:
        """
        This doesn't bind files, but does the equivalent preparation
        of making required files available for the final process to 
        run using them for the barebones container.
        Instead of binding files it creates a symlink of them to a
        specific directory for e4s-cl to find them.
        """
        file_to_bind = Path(path)
        file_basename = file_to_bind.name
        if dest is not None:
            dest = Path(dest)
            if dest.name == 'barebones_script':
                dest.write_text(file_to_bind.read_text())
                os.chmod(dest, 0o755)
            else:
                create_symlink(file_to_bind, Path(dest))
        else:
            create_symlink(file_to_bind, Path(BAREBONES_LIBRARY_DIR) / file_basename)


    def bind_env_var(self, key, value):
        self.env.update({f"{key}": value})

    def _has_nvidia(self):
        # Assume that the proper ldconfig call has been run and that nvidia
        # libraries are listed in the cache
        if 'nvidia' not in " ".join(cache_libraries()):
            LOGGER.debug("Disabling Nvidia support: no libraries found")
            return False
        return True

    def run(self, command: List[str], overload: bool = True) -> int:

        container_cmd = [*self._prepare(command, overload)]

        return run_subprocess(container_cmd, env=self.env)


CLASS = BarebonesContainer



from os import getcwd, environ, pathsep
from unittest import skipIf
from pathlib import Path
import tests
from e4s_cl import config
from e4s_cl.util import which
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
)

CONFIG_EXECUTABLE = tests.ASSETS / 'bin' / 'podman-conf'
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
backends:
  podman:
    executable: '{CONFIG_EXECUTABLE}'
    options: ['--root', '/opt/podman']
    run_options: ['--tz', 'UTC+8']
""")


class ContainerTestPodman(tests.TestCase):

    def test_additional_options_config(self):
        container = Container(name='podman')
        command = ['']

        podman_command = container._prepare(command)
        for option in {'--root', '/opt/podman', '--tz', 'UTC+8'}:
            self.assertNotIn(option, podman_command)

        config.update_configuration(TEST_CONFIGURATION)
        podman_command = container._prepare(command)
        self.assertContainsInOrder([
            '--root',
            '/opt/podman',
            'run',
            '--tz',
            'UTC+8',
        ], podman_command)

        config.update_configuration(DEFAULT_CONFIGURATION)
        podman_command = container._prepare(command)
        for option in {'--root', '/opt/podman', '--tz', 'UTC+8'}:
            self.assertNotIn(option, podman_command)

    def test_additional_options_environment(self):
        container = Container(name='podman')
        command = ['']

        podman_command = container._prepare(command)
        for option in {'--root', '/opt/podman', '--tz', 'UTC+8'}:
            self.assertNotIn(option, podman_command)

        environ['E4S_CL_PODMAN_OPTIONS'] = '--root /opt/podman'
        environ['E4S_CL_PODMAN_RUN_OPTIONS'] = '--tz UTC+8'
        podman_command = container._prepare(command)
        self.assertContainsInOrder([
            '--root',
            '/opt/podman',
            'run',
            '--tz',
            'UTC+8',
        ], podman_command)

        del environ['E4S_CL_PODMAN_OPTIONS']
        del environ['E4S_CL_PODMAN_RUN_OPTIONS']
        podman_command = container._prepare(command)
        for option in {'--root', '/opt/podman', '--tz', 'UTC+8'}:
            self.assertNotIn(option, podman_command)

    def test_executable(self):
        """Assert the default podman executable comes from $PATH"""
        container = Container(name='podman')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"

        self.assertEqual(tests.ASSETS / 'bin' / 'podman',
                         container._executable())

        environ['PATH'] = default_path

    def test_executable_config(self):
        """Assert the podman executable is read from the configuration"""
        container = Container(name='podman')

        config.update_configuration(TEST_CONFIGURATION)
        self.assertEqual(tests.ASSETS / 'bin' / 'podman-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

    def test_executable_env(self):
        """Assert the podman executable is read from the environment"""
        container = Container(name='podman')

        environ['E4S_CL_PODMAN_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                  'podman-env')
        self.assertEqual(tests.ASSETS / 'bin' / 'podman-env',
                         container._executable())

        del environ['E4S_CL_PODMAN_EXECUTABLE']

    def test_executable_priority(self):
        """Assert the environment has precedence over config and config over default"""
        container = Container(name='podman')

        default_path = environ.get('PATH', '')
        environ['PATH'] = f"{tests.ASSETS / 'bin'}{pathsep}{default_path}"
        config.update_configuration(TEST_CONFIGURATION)
        environ['E4S_CL_PODMAN_EXECUTABLE'] = str(tests.ASSETS / 'bin' /
                                                  'podman-env')

        self.assertEqual(tests.ASSETS / 'bin' / 'podman-env',
                         container._executable())

        del environ['E4S_CL_PODMAN_EXECUTABLE']

        self.assertEqual(tests.ASSETS / 'bin' / 'podman-conf',
                         container._executable())

        config.update_configuration(DEFAULT_CONFIGURATION)

        self.assertEqual(tests.ASSETS / 'bin' / 'podman',
                         container._executable())

        environ['PATH'] = default_path



"""
Error-path tests for the 'launch' command (e4s_cl/cli/commands/launch.py).

These tests verify that the command fails gracefully (non-zero exit) when
given incomplete or inconsistent arguments, and that a valid minimal invocation
succeeds in dry-run mode.  No MPI installation, container runtime, or network
access is required.
"""

import tests
from e4s_cl import EXIT_SUCCESS
from e4s_cl.variables import set_dry_run
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.launch import COMMAND


class LaunchErrorPathTest(tests.TestCase):
    """Argument-validation error paths for LaunchCommand.main()."""

    def setUp(self):
        self.resetStorage()
        set_dry_run(False)

    def tearDown(self):
        set_dry_run(False)
        Profile.controller().unselect()
        self.resetStorage()

    # ------------------------------------------------------------------
    # Cases that must fail before any container or MPI call is made
    # ------------------------------------------------------------------

    def test_no_cmd_fails(self):
        """launch with no executable after the options must fail."""
        self.assertNotCommandReturnValue(
            EXIT_SUCCESS, COMMAND,
            ['--backend', 'containerless', '--image', 'some.img'])

    def test_missing_backend_without_profile_fails(self):
        """launch without --backend and no selected profile must fail."""
        self.assertNotCommandReturnValue(
            EXIT_SUCCESS, COMMAND, ['mpirun', '-n', '2', 'hostname'])

    def test_missing_image_without_profile_fails(self):
        """launch with --backend but no --image and no profile must fail."""
        self.assertNotCommandReturnValue(
            EXIT_SUCCESS, COMMAND,
            ['--backend', 'containerless', 'mpirun', 'hostname'])

    # ------------------------------------------------------------------
    # Cases that must succeed in dry-run mode
    # ------------------------------------------------------------------

    def test_barebones_backend_no_image_dry_run(self):
        """'barebones' backend is exempt from the image requirement."""
        set_dry_run(True)
        self.assertCommandReturnValue(
            EXIT_SUCCESS, COMMAND,
            ['--backend', 'barebones', 'mpirun', 'hostname'])

    def test_full_minimal_invocation_dry_run(self):
        """launch with backend + image + launcher + program succeeds in dry-run."""
        set_dry_run(True)
        self.assertCommandReturnValue(
            EXIT_SUCCESS, COMMAND,
            ['--backend', 'containerless', '--image', 'my.sif',
             'mpirun', '-n', '1', 'hostname'])


class LaunchProfileFallbackTest(tests.TestCase):
    """launch reads missing flags from a selected profile."""

    def setUp(self):
        self.resetStorage()
        set_dry_run(True)
        ctrl = Profile.controller()
        profile = ctrl.create({
            'name': 'test_profile',
            'backend': 'containerless',
            'image': 'profile.sif',
        })
        ctrl.select(profile)

    def tearDown(self):
        set_dry_run(False)
        Profile.controller().unselect()
        self.resetStorage()

    def test_launch_uses_profile_backend_and_image(self):
        """Specifying --profile supplies backend and image without extra flags."""
        # Pass --profile explicitly so the test is independent of parser
        # construction order (the parser default is evaluated at first use and
        # may be cached before setUp selects a profile).
        self.assertCommandReturnValue(
            EXIT_SUCCESS, COMMAND,
            ['--profile', 'test_profile', 'mpirun', '-n', '1', 'hostname'])



"""
Module housing support for WI4MPI compatibility
"""

import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple
from pathlib import Path
from functools import lru_cache
from e4s_cl import logger
from e4s_cl.cf.detect_mpi import MPIIdentifier, library_install_dir
from e4s_cl.cf.containers import Container
from e4s_cl.cf.launchers import filter_arguments, Parser
from e4s_cl.cf.launchers.mpirun import _wi4mpi_options

LOGGER = logger.get_logger(__name__)


@dataclass(frozen=True)
class MPIFamily:
    """
    Dataclass to have compact objects with MPI vendor metadata
    """
    # Vendor name, as returned in MPIIdentifier
    vendor_name: str
    # Name used by wi4mpi on the command line
    cli_name: str
    # Name used by wi4mpi in the environment
    env_name: str
    # Wi4MPI's environment variable for this family
    path_key: str
    # Wi4MPI's default environment variable for this family
    default_path_key: str
    # Default soname for this vendor's C library
    mpi_c_soname: str
    # Default soname for this vendor Fortran library
    mpi_f_soname: str

    def __str__(self):
        return self.cli_name


# If the MPI family's specific library name is not found, fallback on the below
# values to possibly avoid a crash
FALLBACK_MPI_C_SONAME = "libmpi.so"
FALLBACK_MPI_F_SONAME = "libmpifort.so"

# MPI vendor libraries metadata. On top of the different MPI families, some
# vendors adopt standard paths different from the norm. This collection keeps
# track of all metadata for each of those vendors.
WI4MPI_METADATA = [
    MPIFamily(
        vendor_name='Intel(R) MPI',
        cli_name='intelmpi',
        env_name='INTEL',
        path_key='INTELMPI_ROOT',
        default_path_key='INTELMPI_DEFAULT_ROOT',
        mpi_c_soname='libmpi.so',
        mpi_f_soname='libmpifort.so',
    ),
    MPIFamily(
        vendor_name='Open MPI',
        cli_name='openmpi',
        env_name='OMPI',
        path_key='OPENMPI_ROOT',
        default_path_key='OPENMPI_DEFAULT_ROOT',
        mpi_c_soname='libmpi.so',
        mpi_f_soname='libmpi_mpifh.so',
    ),
    MPIFamily(
        vendor_name='MPICH',
        cli_name='mpich',
        env_name='MPICH',
        path_key='MPICH_ROOT',
        default_path_key='MPICH_DEFAULT_ROOT',
        mpi_c_soname='libmpi.so',
        mpi_f_soname='libmpifort.so',
    ),
    MPIFamily(
        vendor_name='CRAY MPICH',
        cli_name='mpich',
        env_name='MPICH',
        path_key='MPICH_ROOT',
        default_path_key='MPICH_DEFAULT_ROOT',
        mpi_c_soname='libmpi_cray.so',
        mpi_f_soname='libmpifort_cray.so',
    ),
    MPIFamily(
        vendor_name='MVAPICH',
        cli_name='mvapich',
        env_name='MVAPICH',
        path_key='MVAPICH_ROOT',
        default_path_key='MVAPICH_DEFAULT_ROOT',
        mpi_c_soname='libmpi.so',
        mpi_f_soname='libmpifort.so',
    ),
]

# Keys used to describe MPI families. Wi4MPI is needed if a pair of those keys
# is present in the SUPPORTED_TRANSLATIONS collection
WI4MPI_SOURCES = set(map(lambda x: x.cli_name,
                         WI4MPI_METADATA)) | {'interface'}

SUPPORTED_TRANSLATIONS = {
    ('intelmpi', 'openmpi'),
    ('interface', 'intelpmi'),
    ('interface', 'mpich'),
    ('interface', 'mvapich'),
    ('interface', 'openmpi'),
    ('mpich', 'openmpi'),
    ('mvapich', 'openmpi'),
    ('openmpi', 'intelpmi'),
    ('openmpi', 'mpich'),
    ('openmpi', 'openmpi'),
    ('intelmpi', 'mvapich'),
    ('openmpi', 'mvapich'),
}

_FAMILY_ENV_VARS = set(map(lambda x: x.path_key, WI4MPI_METADATA))

# Set of all environment variables used by Wi4MPI, to pass to the underlying code
WI4MPI_ENVIRONMENT_VARIABLES = {
    'WI4MPI_ROOT',
    'WI4MPI_FROM',
    'WI4MPI_TO',
    'WI4MPI_RUN_MPI_C_LIB',
    'WI4MPI_RUN_MPI_F_LIB',
    'WI4MPI_RUN_MPIIO_C_LIB',
    'WI4MPI_RUN_MPIIO_F_LIB',
} | _FAMILY_ENV_VARS


def wi4mpi_identify(value: str) -> Optional[MPIFamily]:
    for family in WI4MPI_METADATA:
        if value.lower() in {
                family.vendor_name.lower(),
                family.cli_name.lower()
        }:
            return family
    return None


def wi4mpi_qualifier(value: MPIIdentifier) -> Optional[str]:
    if not isinstance(value, MPIIdentifier):
        return None
    match = wi4mpi_identify(value.vendor)
    if match:
        return match.cli_name
    return None


def wi4mpi_get_metadata(value: MPIIdentifier) -> Optional[MPIFamily]:
    if not isinstance(value, MPIIdentifier):
        return None
    match = wi4mpi_identify(value.vendor)
    if match:
        return match
    return None


def wi4mpi_enabled() -> bool:
    # Convoluted way to have True or False instead of the contents of the key
    return not os.environ.get("WI4MPI_VERSION") is None


@lru_cache()
def wi4mpi_root() -> Optional[Path]:
    string = os.environ.get("WI4MPI_ROOT")

    if string is None or not string:
        LOGGER.debug("Getting Wi4MPI root failed - WI4MPI_ROOT not in environment")
        LOGGER.debug(
            "Available Wi4MPI env vars: WI4MPI_FROM=%s, WI4MPI_TO=%s",
            os.environ.get("WI4MPI_FROM"),
            os.environ.get("WI4MPI_TO"),
        )
        return None

    LOGGER.debug("Wi4MPI root found: %s", string)
    return Path(string)


def __read_cfg(cfg_file: Path) -> Dict[str, str]:
    config = {}

    try:
        with open(cfg_file, 'r', encoding='utf-8') as cfg:
            for line in cfg.readlines():
                line = line.strip()

                if line.startswith('#') or not '=' in line:
                    continue

                key, value = line.split('=')

                config.update({key: value.strip('"')})

    except OSError as err:
        LOGGER.debug("Error accessing configuration %s: %s",
                     cfg_file.as_posix(), str(err))

    return config


@lru_cache()
def wi4mpi_config(install_dir: Path) -> Dict[str, str]:
    global_cfg = __read_cfg(install_dir / 'etc' / 'wi4mpi.cfg')
    user_cfg = __read_cfg(Path.home() / '.wi4mpi.cfg')

    global_cfg.update(user_cfg)

    return global_cfg


def wi4mpi_import(container: Container, install_dir: Path) -> None:
    """
    Bind to a container the necessary files required for wi4mpi to run
    """
    container.bind_file(install_dir.as_posix())

    config = wi4mpi_config(install_dir)

    for (key, value) in config.items():
        if 'ROOT' in key and value:
            container.bind_file(value)
            container.add_ld_library_path(
                Path(value).joinpath('lib').as_posix())


def wi4mpi_libraries(install_dir: Path) -> List[Path]:
    """
    Use the environment to output a list of libraries required by wi4mpi
    """
    config = wi4mpi_config(install_dir)

    source = os.environ.get("WI4MPI_FROM", "")
    target = os.environ.get("WI4MPI_TO", "")

    if not (source and target):
        LOGGER.debug(
            "Error getting WI4MPI libraries: Missing environment variables")
        return []

    wrapper_lib = Path(install_dir, 'libexec', 'wi4mpi',
                       f"libwi4mpi_{source}_{target}.so")

    def _get_lib(env_name: str) -> Optional[Path]:
        matches = set(filter(lambda x: x.env_name == env_name,
                             WI4MPI_METADATA))
        if len(matches) == 1:
            distribution_data = matches.pop()

            config_value = config.get(distribution_data.default_path_key, "")

            return Path(config_value, 'lib', 'libmpi.so')
        return None

    source_lib = _get_lib(source)
    target_lib = _get_lib(target)

    return list(
        filter(lambda x: x.resolve().exists(),
               filter(None, [wrapper_lib, source_lib, target_lib])))


def wi4mpi_libpath(install_dir: Path):
    """
    Select all WI4MPI-relevant elements from the LD_LIBRARY_PATH
    """
    ld_library_path = os.environ.get('LD_LIBRARY_PATH', '').split(':')

    for filename in ld_library_path:
        if install_dir.as_posix() in filename:
            yield Path(filename)


def wi4mpi_preload(install_dir: Path, container_library_dir: str = None) -> List[str]:
    """
    Returns a list of libraries to preload for WI4MPI
    
    Preload order follows Wi4MPI's official script:
    1. The Wi4MPI translation library (libwi4mpi_{FROM}_{TO}.so) - intercepts MPI calls
    2. The target MPI Fortran library (WI4MPI_RUN_MPI_F_LIB) - provides symbols for translation
    3. The target MPI C library (WI4MPI_RUN_MPI_C_LIB) - provides symbols like ompi_mpi_char
    
    NOTE: The fake libraries (fakelib{FROM}/) go in LD_LIBRARY_PATH, NOT in LD_PRELOAD.
    This matches the official Wi4MPI launcher script behavior.
    
    IMPORTANT: When container_library_dir is provided, this function assumes the target
    MPI libraries (WI4MPI_RUN_MPI_C_LIB and WI4MPI_RUN_MPI_F_LIB) have already been
    imported into the container at that location. The caller must ensure import_library()
    is called before this function when running in container mode.
    
    Args:
        install_dir: Path to Wi4MPI installation directory
        container_library_dir: The path where host libraries are mounted inside
            the container (e.g., '/.e4s-cl/hostlibs'). Required for container execution.
    """
    to_preload = []

    source = os.environ.get("WI4MPI_FROM", "")
    target = os.environ.get("WI4MPI_TO", "")

    # 1. Add the translation library FIRST
    if source and target:
        translation_lib = install_dir / 'libexec' / 'wi4mpi' / f"libwi4mpi_{source}_{target}.so"
        if translation_lib.exists():
            to_preload.append(translation_lib.as_posix())
            LOGGER.debug("Wi4MPI: Adding translation library to preload: %s", translation_lib)
        else:
            LOGGER.warning("Wi4MPI translation library not found: %s", translation_lib)

    # 2. Add the target MPI F library (provides Fortran symbols)
    target_mpi_f_lib = os.environ.get("WI4MPI_RUN_MPI_F_LIB", "")
    if target_mpi_f_lib and Path(target_mpi_f_lib).exists():
        if container_library_dir:
            container_path = Path(container_library_dir) / Path(target_mpi_f_lib).name
            to_preload.append(container_path.as_posix())
            LOGGER.debug("Wi4MPI: Adding target MPI F library to preload (container path): %s", container_path)
        else:
            to_preload.append(target_mpi_f_lib)
            LOGGER.debug("Wi4MPI: Adding target MPI F library to preload: %s", target_mpi_f_lib)

    # 3. Add the target MPI C library (provides C symbols like ompi_mpi_char)
    target_mpi_c_lib = os.environ.get("WI4MPI_RUN_MPI_C_LIB", "")
    if target_mpi_c_lib and Path(target_mpi_c_lib).exists():
        if container_library_dir:
            container_path = Path(container_library_dir) / Path(target_mpi_c_lib).name
            to_preload.append(container_path.as_posix())
            LOGGER.debug("Wi4MPI: Adding target MPI C library to preload (container path): %s", container_path)
        else:
            to_preload.append(target_mpi_c_lib)
            LOGGER.debug("Wi4MPI: Adding target MPI C library to preload: %s", target_mpi_c_lib)

    # NOTE: Fake libraries are NOT preloaded - they go in LD_LIBRARY_PATH
    # (handled by wi4mpi_libpath function)

    return to_preload


def wi4mpi_adapt_arguments(cmd_line: List[str]) -> List[str]:
    """Quote arguments destined to the implementation mpirun in an --extra block"""
    launcher = None
    arguments = cmd_line
    if cmd_line[0] == 'mpirun':
        launcher = 'mpirun'
        arguments = cmd_line[1:]

    wi4mpi, mpi = filter_arguments(Parser(_wi4mpi_options), arguments)
    extra = []
    if mpi:
        extra = ['-E', " ".join(mpi)]

    return list(filter(None, [launcher, *wi4mpi, *extra]))


def wi4mpi_find_libraries(
        target: MPIFamily,
        mpi_libraries: List[Path]) -> Tuple[Optional[Path], Optional[Path]]:
    """
    Find the target's required MPI libraries from the paths in mpi_libraries
    """

    def locate(soname: str, available: List[Path]) -> Optional[Path]:
        """
        Find the library with the given soname, either from the libraries
        passed in the list or from the directories they are in
        """
        matches = set(filter(lambda x: x.name.startswith(soname), available))
        search_directories = set(map(lambda x: x.resolve().parent, available))

        # If a match exists in the given libraries
        if matches:
            return matches.pop()

        # Search the libraries' directories for the soname
        for directory in search_directories:
            matches = set(directory.glob(f"{soname}*"))
            if matches:
                return matches.pop()

        LOGGER.debug(
            "Failed to locate %(soname)s in %(directories)s",
            {
                "soname": soname,
                "directories": search_directories,
            },
        )

        return None

    # Find the entry C and Fortran MPI libraries
    run_c_lib = (locate(target.mpi_c_soname, mpi_libraries)
                 or locate(FALLBACK_MPI_C_SONAME, mpi_libraries))
    run_f_lib = (locate(target.mpi_f_soname, mpi_libraries)
                 or locate(FALLBACK_MPI_F_SONAME, mpi_libraries))

    return (run_c_lib, run_f_lib)


def wi4mpi_prepare_environment_preload(
    wi4mpi_install_dir: Path,
    from_: str,
    target: MPIFamily,
    target_install_dir: Path,
    run_c_lib: Path,
    run_f_lib: Path,
):
    """
    Prepare the environment for a Wi4MPI execution in preload mode
    Updates the environment variables:
      + WI4MPI_ROOT
      + WI4MPI_FROM
      + WI4MPI_TO
      + <LIBRARY>_ROOT
      + WI4MPI_RUN_MPI_C_LIB
      + WI4MPI_RUN_MPI_F_LIB
      + WI4MPI_RUN_MPIIO_C_LIB
      + WI4MPI_RUN_MPIIO_F_LIB
    """

    env = {
        'WI4MPI_ROOT': str(wi4mpi_install_dir),
        target.path_key: str(target_install_dir),
        'WI4MPI_FROM': str(from_),
        'WI4MPI_TO': str(target.cli_name),
        'WI4MPI_RUN_MPI_C_LIB': str(run_c_lib),
        'WI4MPI_RUN_MPI_F_LIB': str(run_f_lib),
        'WI4MPI_RUN_MPIIO_C_LIB': str(run_c_lib),
        'WI4MPI_RUN_MPIIO_F_LIB': str(run_f_lib),
    }

    for key, value in env.items():
        LOGGER.debug("Wi4MPI preload: %s=%s", key, value)
        os.environ[key] = value


def wi4mpi_prepare_environment_interface(
    wi4mpi_install_dir: Path,
    target: MPIFamily,
    target_install_dir: Path,
    run_c_lib: Path,
    run_f_lib: Path,
):
    """
    Prepare the environment for a Wi4MPI execution in interface mode
    Updates the environment variables:
      + WI4MPI_ROOT
      + WI4MPI_TO
      + <LIBRARY>_ROOT
      + WI4MPI_RUN_MPI_C_LIB
      + WI4MPI_RUN_MPI_F_LIB
      + WI4MPI_RUN_MPIIO_C_LIB
      + WI4MPI_RUN_MPIIO_F_LIB
    """

    env = {
        'WI4MPI_ROOT': str(wi4mpi_install_dir),
        target.path_key: str(target_install_dir),
        'WI4MPI_TO': str(target.cli_name),
        'WI4MPI_RUN_MPI_C_LIB': str(run_c_lib),
        'WI4MPI_RUN_MPI_F_LIB': str(run_f_lib),
        'WI4MPI_RUN_MPIIO_C_LIB': str(run_c_lib),
        'WI4MPI_RUN_MPIIO_F_LIB': str(run_f_lib),
    }

    for key, value in env.items():
        LOGGER.debug("Wi4MPI interface: %s=%s", key, value)
        os.environ[key] = value



"""
Entry point to the CLI
"""

import os
import sys
import e4s_cl
from e4s_cl import (
    E4S_CL_SCRIPT,
    E4S_CL_VERSION,
    PYTHON_VERSION,
    cli,
    config,
    logger,
    util,
)
from e4s_cl.variables import DryRunAction
from e4s_cl.cf.version import Version
from e4s_cl.cli import UnknownCommandError, arguments
from e4s_cl.cli.command import AbstractCommand
from e4s_cl.cli.commands.launch import COMMAND as LAUNCH_COMMAND
from e4s_cl.util import sanitize_args

LOGGER = logger.get_logger(__name__)

HELP_PAGE_FMT = """'%(command)s' page to be written."""


class MainCommand(AbstractCommand):
    """Main entry point to the command line interface."""

    def __init__(self):
        summary_parts = [
            util.color_text(f"E4S Container Launcher {E4S_CL_VERSION}",
                            'red',
                            attrs=['bold'])
        ]

        super().__init__(__name__,
                         summary_fmt=''.join(summary_parts),
                         help_page_fmt=HELP_PAGE_FMT)

        self.command = os.path.basename(E4S_CL_SCRIPT)

    def _construct_parser(self):
        usage = f"{self.command} [arguments] <subcommand> [options]"
        epilog_parts = [
            "",
            cli.commands_description(), "",
            "See `%(command)s help <subcommand>` for more information on a subcommand."
        ]
        epilog = '\n'.join(epilog_parts) % {
            'color_command': util.color_text(self.command, 'cyan'),
            'command': self.command
        }
        parser = arguments.get_parser(
            prog=self.command,
            usage=usage,
            description=self.summary,
            epilog=epilog,
        )

        parser.add_argument(
            'command',
            help="See subcommand descriptions below",
            choices=cli.commands_next(),
            metavar='<subcommand>',
        )

        parser.add_argument(
            'options',
            help="Options to be passed to <subcommand>",
            metavar='[options]',
            nargs=arguments.REMAINDER,
        )

        parser.add_argument(
            '-d',
            '--dry-run',
            nargs=0,
            help="Do nothing, print out what would be done instead",
            action=DryRunAction,
        )

        verbosity_group = parser.add_mutually_exclusive_group()

        verbosity_group.add_argument(
            '-v',
            '--verbose',
            help="show debugging messages",
            const='DEBUG',
            default=arguments.SUPPRESS,
            action='store_const',
        )

        verbosity_group.add_argument(
            '-q',
            '--quiet',
            help="suppress all output except error messages",
            const='ERROR',
            default=arguments.SUPPRESS,
            action='store_const',
        )

        parser.add_argument(
            '-V',
            '--version',
            action='version',
            version=e4s_cl.version_banner(),
        )

        parser.add_argument(
            '--print-config',
            help="Print a template of configuration file",
            action='version',
            version=config.ALLOWED_CONFIG.template(),
        )

        return parser

    def set_log_level(self, argList):
        quiet = getattr(argList, 'quiet', logger.LOG_LEVEL)
        verbose = getattr(argList, 'verbose', quiet)
        logger.set_log_level(verbose)

    def _py38_parse(self, argv):
        """
        If a command is not detected, insert 'launch' in the argv array
        """

        # Debug is not enabled until below, so the following statement
        # enables it for this code block
        self.set_log_level(argv)

        empty = len(argv) == 0
        command = set(argv) & set(cli.commands_next())

        # Get a list of valid option strings from the parser
        option_strings = util.flatten(
            map(lambda x: x.option_strings, self.parser.actions))

        # If the error is not related to the omission of subcommand
        if not empty and not command:
            # Insert `launch` after any valid option string for e4s-cl
            for arg in argv:
                if arg in option_strings:
                    continue
                argv.insert(argv.index(arg), LAUNCH_COMMAND.monicker)
                break

        LOGGER.debug("Parsing updated arguments '%s'", argv)

        return self._parse_args(argv)

    def _py39_parse(self, argv):
        """
        Use the exit_on_error from python 3.9 to validate a command line,
        and complete with 'launch' in case it misses a sub-command
        """

        # Disable built-in error catching for this special case
        self.parser.exit_on_error = False

        try:
            args = self._parse_args(argv)
        except arguments.ArgumentError as err:
            # Debug is not enabled until below, so the following statement
            # enables it for this code block
            self.set_log_level(argv)

            LOGGER.debug("Argument parsing errored out with '%s'", argv)

            # Check for the presence of a e4s-cl command
            empty = len(argv) == 0
            command = set(argv) & set(cli.commands_next())

            # If the error is not related to the omission of subcommand
            if empty or command:
                # Snippet coming from Lib/argparse.py:1853
                err = sys.exc_info()[1]
                self.parser.error(str(err))

            # Disable error catching
            self.parser.exit_on_error = True

            # Get a list of valid option strings from the parser
            option_strings = util.flatten(
                map(lambda x: x.option_strings, self.parser.actions))

            # Insert `launch` after any valid option string for e4s-cl
            for arg in argv:
                if arg in option_strings:
                    continue
                argv.insert(argv.index(arg), LAUNCH_COMMAND.monicker)
                break

            LOGGER.debug("Parsing updated arguments '%s'", argv)
            args = self._parse_args(argv)
        finally:
            self.parser.exit_on_error = True

        return args

    def main(self, argv):
        """Program entry point.

        Args:
            argv (list): Command line arguments.

        Returns:
            int: Process return code: non-zero if a problem occurred, 0 otherwise
        """

        if Version('.'.join(map(str, PYTHON_VERSION))) < Version('3.9.0'):
            parse_method = self._py38_parse
        else:
            parse_method = self._py39_parse

        argv = sanitize_args(argv)
        args = parse_method(argv)

        self.set_log_level(args)

        # Try to execute as a command
        try:
            cmd = args.command
            cmd_args = args.options

            return cli.execute_command([cmd], cmd_args)
        except UnknownCommandError:
            pass

        # Not sure what to do at this point, so advise the user and exit
        LOGGER.info("Unknown command.  Calling `%s help %s` to get advice.",
                    E4S_CL_SCRIPT, cmd)

        return cli.execute_command(['help'], [cmd])


COMMAND = MainCommand()

if __name__ == '__main__':
    sys.exit(COMMAND.main(sys.argv[1:]))
